#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

import subprocess
from enum import Enum
from pathlib import Path
from typing import Any, Sequence, Union

import numpy
from cycler import Cycler
from matplotlib import patches, pyplot, rcParams
from matplotlib.legend_handler import HandlerErrorbar
from sorcery import assigned_names

from draugr.visualisation.matplotlib_utilities.quirks import auto_post_hatch
from draugr.visualisation.matplotlib_utilities.styles.annotation import (
    rt_ann_transform,
    semi_opaque_round_tight_bbox,
)
from draugr.visualisation.matplotlib_utilities.styles.cyclers import (
    monochrome_hatch_cycler,
    monochrome_line_cycler,
    simple_hatch_cycler,
)

__all__ = [
    "denormalise_minusoneone",
    "matplotlib_bounding_box",
    "remove_decoration",
    "use_monochrome_style",
    "decolorise_plot",
    "save_embed_fig",
    "latex_clean_label",
    "make_errorbar_legend",
    "annotate_point",
    "MatplotlibHorizontalAlignment",
    "MatplotlibVerticalAlignment",
]

from matplotlib.patches import Rectangle
from matplotlib.pyplot import legend

from warg import Number, passes_kws_to

# Cycler
##Addition
###Equal length Cycler s with different keys can be added to get the ‘inner’ product of two cycles
##Multiplication
###which gives the ‘outer product’ of the two cycles (same as itertools.prod() )

from matplotlib.axes import Axes, ErrorbarContainer


class MatplotlibHorizontalAlignment(Enum):
    """ """

    (center, right, left) = assigned_names()


class MatplotlibVerticalAlignment(Enum):
    """ """

    (center, top, bottom, baseline, center_baseline) = assigned_names()


def latex_clean_label(s: str) -> str:
    """Clean label for troublesome symbols"""
    return s.replace("_", " ")  # .replace('\\',' ').replace('/',' ')


def make_errorbar_legend(ax: Axes = None) -> None:
    """size adjusted legend for error bars, default does not work well with different linestyles"""

    if ax is None:
        ax = pyplot.gca()
    ax.legend(
        handlelength=5,
        handleheight=3,
        handler_map={ErrorbarContainer: HandlerErrorbar(xerr_size=0.9, yerr_size=0.9)},
    )


def annotate_point(ax: Axes, x: Sequence, y: Sequence, t: Any) -> None:
    """

    :param ax:
    :param x:
    :param y:
    :param t:
    :return:
    """
    if ax is None:
        ax = pyplot.gca()
    ax.annotate(
        f"{t:.2f}",
        (x, y),
        textcoords="offset points",
        fontsize="xx-small",
        bbox=semi_opaque_round_tight_bbox,
        annotation_clip=True,
        # see details https://github.com/matplotlib/matplotlib/issues/14354#issuecomment-523630316
        clip_on=True,
        **rt_ann_transform,
    )


@passes_kws_to(pyplot.savefig)
def save_embed_fig(
    path: Union[Path, str],
    bbox_inches: Union[Number, str] = "tight",
    transparent: bool = True,
    attempt_fix_empty_white_space: bool = False,
    post_process_crop: bool = False,
    ax: Axes = None,
    suffix: str = ".pdf",
    **kwargs,
) -> None:
    """Save fig for latex pdf embedding"""

    if attempt_fix_empty_white_space:  # remove it
        # pyplot.gca().set_axis_off()
        # pyplot.axis('off') # this rows the rectangular frame
        # ax.get_xaxis().set_visible(False) # this removes the ticks and numbers for x axis
        # ax.get_yaxis().set_visible(False) # this removes the ticks and numbers for y axis
        if ax is None:
            ax = pyplot.gca()
        pyplot.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        pyplot.margins(0, 0)
        ax.xaxis.set_major_locator(pyplot.NullLocator())
        ax.yaxis.set_major_locator(pyplot.NullLocator())

    """
clip_box = Bbox(((0,0),(300,300)))
for o in pyplot.findobj():
o.set_clip_on(True)
o.set_clip_box(clip_box)

"""

    if not isinstance(path, Path):
        path = Path(path)

    path_str = str(path.with_suffix(suffix))
    pyplot.savefig(
        path_str,
        bbox_inches=bbox_inches,
        transparent=transparent,
        **kwargs,
    )

    if (
        post_process_crop and suffix == ".pdf"
    ):  # Generally a good idea since matplotlib does not exclude invisible parts(eg. data points or anchors) of the plot.
        from pdfCropMargins import crop  # pip install pdfCropMargins

        crop(
            [
                "-p",
                "0",  # remove all percentage of old margin
                "-a",
                "-5",  # add 5bp margin around all
                path_str,
            ]
        )


def remove_decoration(ax: Axes) -> None:
    """ """
    transparent = (1.0, 1.0, 1.0, 0.0)

    ax.w_xaxis.set_pane_color(transparent)
    ax.w_yaxis.set_pane_color(transparent)
    ax.w_zaxis.set_pane_color(transparent)

    ax.w_xaxis.line.set_color(transparent)
    ax.w_yaxis.line.set_color(transparent)
    ax.w_zaxis.line.set_color(transparent)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])


def denormalise_minusoneone(
    t: Union[numpy.ndarray, Number], coordinates: Union[numpy.ndarray, Number]
) -> Union[numpy.ndarray, Number]:
    """ """
    return 0.5 * ((coordinates + 1.0) * t)


def decolorise_plot(ax_: Axes, inverted: bool = False) -> None:
    """

    set black and white edge colors and face colors respectively. Converse if inverted.
    """
    edge_color = "k"
    face_color = "w"
    if inverted:
        face_color, edge_color = edge_color, face_color
    pyplot.setp(ax_.artists, edgecolor=edge_color, facecolor=face_color)
    pyplot.setp(ax_.lines, color=edge_color)


def matplotlib_bounding_box(
    x: Number, y: Number, size: Number, color: str = "w"
) -> Rectangle:
    """ """
    x = int(x - (size / 2))
    y = int(y - (size / 2))
    rect = patches.Rectangle(
        (x, y), size, size, linewidth=1, edgecolor=color, fill=False
    )
    return rect


def use_monochrome_style(
    prop_cycler: Cycler = monochrome_line_cycler,
    # ONLY COLOR AND LINESTYLE MAKES SENSE FOR NOW, matplotlib seems very undone in this api atleast for bars
) -> None:
    """

    :param prop_cycler:
    """
    # from matplotlib.pyplot import axes, grid

    # from matplotlib import lines, markers
    # print(lines.lineStyles.keys(),markers.MarkerStyle.markers.keys())

    # print(rcParams.keys())

    # pyplot.style.use('default')
    # print(pyplot.style.available)
    # print(matplotlib.matplotlib_fname())

    pyplot.style.use(Path(__file__).parent / "styles" / "monochrome.mplstyle")
    if pyplot.rcParams["text.usetex"]:
        try:
            report = subprocess.check_output("latex -v", stderr=subprocess.STDOUT)
        except FileNotFoundError as exc:
            msg = f'No tex: {"latex"}'
            # raise RuntimeError(msg)
            print(f"{msg}, disabling")
            pyplot.rcParams.update({"text.usetex": False})
    rcParams.update({"axes.prop_cycle": prop_cycler})
    # auto_post_print_dpi()


if __name__ == "__main__":

    def asiuhda() -> None:
        """
        :rtype: None
        """
        use_monochrome_style()
        bar_styles = monochrome_hatch_cycler()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.bar(x, numpy.random.randint(2, 10), **next(bar_styles), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    def asiuh214da() -> None:
        """
        :rtype: None
        """
        pyplot.style.use(Path(__file__).parent / "styles" / "monochrome.mplstyle")
        line_styles = monochrome_line_cycler()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.plot(numpy.random.rand(10), **next(line_styles), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    def asiuasdashda() -> None:
        """
        :rtype: None
        """
        use_monochrome_style(prop_cycler=monochrome_hatch_cycler)

        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.bar(x, numpy.random.randint(2, 10), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        auto_post_hatch(ax, simple_hatch_cycler)
        pyplot.show()

    def asiuhdsada_non() -> None:
        """
        :rtype: None
        """
        fig, ax = pyplot.subplots(1, 1)
        for x in range(3):
            ax.plot(numpy.random.rand(10), label=f"{x}")

        legend()
        pyplot.show()

    def asiuhd21412sada() -> None:
        """
        :rtype: None
        """
        use_monochrome_style()
        fig, ax = pyplot.subplots(1, 1)

        for x in range(3):
            ax.plot(numpy.random.rand(10), label=f"{x}")

        legend()
        from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines

        fix_edge_gridlines(ax)
        pyplot.show()

    # asiuhda()
    # asiuasdashda()
    asiuhd21412sada()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

from enum import Enum
from typing import Tuple

import numpy
from matplotlib import patheffects, pyplot

__all__ = [
    "plot_median_labels",
    "show_values_on_bars",
    "VisualisationErrorStyle",
]

from draugr.visualisation.matplotlib_utilities.styles.annotation import (
    semi_opaque_round_tight_bbox,
)


class VisualisationErrorStyle(Enum):
    band = "band"
    bar = "bars"


def plot_median_labels(
    ax: pyplot.Axes,
    *,
    has_fliers: bool = False,
    # text_size: int = 10,
    # text_weight: str = "normal",
    stroke_width: int = 0,
    precision: int = 3,
    color: str = "black",
    edgecolor: str = "black",  # also the stroke color
    ha: str = "center",
    va: str = "center",  # bottom
    bbox: Tuple = semi_opaque_round_tight_bbox,
) -> None:
    """ """
    lines = ax.get_lines()
    # depending on fliers, toggle between 5 and 6 lines per box
    lines_per_box = 5 + int(has_fliers)
    # iterate directly over all median lines, with an interval of lines_per_box
    # this enables labeling of grouped data without relying on tick positions
    for median_line in lines[4 : len(lines) : lines_per_box]:
        # get center of median line
        mean_x = sum(median_line._x) / len(median_line._x)
        mean_y = sum(median_line._y) / len(median_line._y)

        text = ax.text(
            mean_x,
            mean_y,
            f"{round(mean_y, precision)}",
            ha=ha,
            va=va,
            # fontweight=text_weight,
            # size=text_size,
            color=color,
            # edgecolor=edgecolor
            bbox=bbox,
        )  # print text to center coordinates

        if stroke_width:
            # create small black border around white text
            # for better readability on multi-colored boxes
            text.set_path_effects(
                [
                    patheffects.Stroke(linewidth=stroke_width, foreground=edgecolor),
                    patheffects.Normal(),
                ]
            )


def show_values_on_bars(axs: pyplot.Axes, h_v: str = "v", space: float = 0.4) -> None:
    """ """

    def _show_on_single_plot(ax):
        if h_v == "v":
            for p in ax.patches:
                _x = p.get_x() + p.get_width() / 2
                _y = p.get_y() + p.get_height()
                value = int(p.get_height())
                ax.text(_x, _y, value, ha="center")
        elif h_v == "h":
            for p in ax.patches:
                _x = p.get_x() + p.get_width() + float(space)
                _y = p.get_y() + p.get_height()
                value = int(p.get_width())
                ax.text(_x, _y, value, ha="left")

    if isinstance(axs, numpy.ndarray):
        for idx, ax in numpy.ndenumerate(axs):
            _show_on_single_plot(ax)
    else:
        _show_on_single_plot(axs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 27/06/2020
           """

__all__ = ["get_num_parameters"]

from torch import nn

from draugr.torch_utilities.optimisation.parameters.trainable import (
    trainable_parameters,
)


def get_num_parameters(model: nn.Module, *, only_trainable: bool = False) -> int:
    """
    Returns the number of parameters of a model (trainable or all)

    :param only_trainable:
    :type only_trainable:
    :param model:
    :type model:
    :return:
    :rtype:"""
    if only_trainable:
        return sum(p.numel() for p in trainable_parameters(model))
    return sum(param.numel() for param in model.parameters())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/08/2020
           """

__all__ = ["GtkThemePreferences"]

from typing import Union

from draugr.os_utilities.linux_utilities.gtk_utilities.gtk_settings import GtkSettings


class GtkThemePreferences(GtkSettings):
    """
    Presents a slim series of properties for manipulation of GTK settings
    """

    @property
    def theme(self) -> Union[str, bytes]:
        """

        :return:
        """
        return self.settings.get_property("gtk-theme-name")

    @theme.setter
    def theme(self, theme_name: str) -> None:
        self.settings.set_property("gtk-theme-name", theme_name)

    @property
    def prefer_dark_mode(self) -> bool:
        """

        :return:
        """
        return self.settings.get_property("gtk-application-prefer-dark-theme")

    @prefer_dark_mode.setter
    def prefer_dark_mode(self, enabled: bool) -> None:
        self.settings.set_property("gtk-application-prefer-dark-theme", enabled)


if __name__ == "__main__":

    def asdad() -> None:
        """
        :rtype: None
        """
        a = GtkThemePreferences()

        print(a.prefer_dark_mode)
        a.prefer_dark_mode = not a.prefer_dark_mode
        print(a.prefer_dark_mode)
        a.prefer_dark_mode = not a.prefer_dark_mode
        print(a.prefer_dark_mode)

    def asdad2312() -> None:
        """
        :rtype: None
        """
        with GtkThemePreferences() as a:
            print(a.prefer_dark_mode)

        with GtkThemePreferences() as a:
            print(a.prefer_dark_mode)
            a.prefer_dark_mode = not a.prefer_dark_mode
            print(a.prefer_dark_mode)

        with GtkThemePreferences() as a:
            print(a.prefer_dark_mode)

    def asda213sad2312d() -> None:
        """
        :rtype: None
        """
        a = GtkThemePreferences()

        print(a.prefer_dark_mode)
        a.prefer_dark_mode = not a.prefer_dark_mode  # TODO: DOES NOT WORK!
        print(a.prefer_dark_mode)

    asda213sad2312d()
    # asdad2312()
    # asdad()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-05-2021
           """

import pandas

__all__ = [
    "nested_dict_to_two_level_column_df",
    "nested_dict_to_four_level_column_df",
    "nested_dict_to_three_level_column_df",
]


def nested_dict_to_two_level_column_df(df):  # TODO: MAKE ANY LEVEL!
    """

    :param df:
    :return:
    """
    cols, data = [], []
    index = None
    for k, v in df.items():
        if not isinstance(v, dict):
            cols.append((k, 0))
            data.append(v)
        else:
            for k2, v2 in v.items():
                cols.append((k, k2))
                data.append(v2)
                if index is None:
                    index = v2.index

    return pandas.DataFrame(
        list(zip(*data)), columns=pandas.MultiIndex.from_tuples(cols), index=index
    )


def nested_dict_to_three_level_column_df(df):  # TODO: MAKE ANY LEVEL!
    """

    :param df:
    :return:
    """
    cols, data = [], []
    index = None
    for k, v in df.items():
        for k2, v2 in v.items():
            for k3, v3 in v2.items():
                cols.append((k, k2, k3))
                data.append(v3)
                if index is None:
                    index = v3.index

    return pandas.DataFrame(
        list(zip(*data)), columns=pandas.MultiIndex.from_tuples(cols), index=index
    )


def nested_dict_to_four_level_column_df(df):  # TODO: MAKE ANY LEVEL!
    """

    :param df:
    :return:
    """
    cols, data = [], []
    index = None
    for k, v in df.items():
        for k2, v2 in v.items():
            for k3, v3 in v2.items():
                for k4, v4 in v3.items():
                    cols.append((k, k2, k3, k4))
                    data.append(v4)
                    if index is None:
                        index = v4.index

    return pandas.DataFrame(
        list(zip(*data)), columns=pandas.MultiIndex.from_tuples(cols), index=index
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "Christian Heider Nielsen"
__doc__ = ""

import torch
from torch import nn
from torch.nn import functional

from draugr.torch_utilities.architectures.mlp import MLP


class RecurrentCategoricalMLP(MLP):
    def __init__(self, r_hidden_layers=10, **kwargs):
        super().__init__(**kwargs)
        self._r_hidden_layers = r_hidden_layers
        self._r_input_shape = self._output_shape + r_hidden_layers

        self.hidden = nn.Linear(
            self._r_input_shape, r_hidden_layers, bias=self._use_bias
        )
        self.out = nn.Linear(self._r_input_shape, r_hidden_layers, bias=self._use_bias)

        self._prev_hidden_x = torch.zeros(r_hidden_layers)

    def forward(self, x, **kwargs):
        x = super().forward(x, **kwargs)
        combined = torch.cat((x, self._prev_hidden_x), 1)
        out_x = self.out(combined)
        hidden_x = self.hidden(combined)
        self._prev_hidden_x = hidden_x

        return functional.log_softmax(out_x, dim=-1)


class ExposedRecurrentCategoricalMLP(RecurrentCategoricalMLP):
    def forward(self, x, hidden_x, **kwargs):
        self._prev_hidden_x = hidden_x
        out_x = super().forward(x, **kwargs)

        return functional.log_softmax(out_x, dim=-1), self._prev_hidden_x


class RecurrentBase(nn.Module):
    def __init__(self, recurrent, recurrent_input_size, hidden_size):
        super().__init__()

        self._hidden_size = hidden_size
        self._recurrent = recurrent

        if recurrent:
            self.gru = nn.GRUCell(recurrent_input_size, hidden_size)
            nn.init.orthogonal_(self.gru.weight_ih.data)
            nn.init.orthogonal_(self.gru.weight_hh.data)
            self.gru.bias_ih.data.fill_(0)
            self.gru.bias_hh.data.fill_(0)

    def _forward_gru(self, x, hxs, masks):
        if x.size(0) == hxs.size(0):
            x = hxs = self.gru(x, hxs * masks)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.reshape(T, N, x.size(1))

            # Same deal with masks
            masks = masks.reshape(T, N, 1)

            outputs = []
            for i in range(T):
                hx = hxs = self.gru(x[i], hxs * masks[i])
                outputs.append(hx)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.stack(outputs, dim=0)
            # flatten
            x = x.reshape(T * N, -1)

        return x, hxs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-12-2020
           """

import numpy

__all__ = [
    "normalise_signal",
    "normalise_signal_minmax",
    "normalise_signal_max_abs",
    "heaviside",
]

from draugr.numpy_utilities.signal_utilities.signal_statistics import mean_square


def normalise_signal_minmax(signal: numpy.ndarray) -> numpy.ndarray:
    """ """
    return numpy.interp(signal, (signal.min(), signal.max()), (-1, 1))


def normalise_signal_max_abs(signal: numpy.ndarray) -> numpy.ndarray:
    """ """
    return signal / numpy.abs(signal).max()


def normalise_signal(
    signal: numpy.ndarray, variance: numpy.ndarray = None
) -> numpy.ndarray:
    """Normalise power in y to a (standard normal) white noise signal.
    Optionally normalise to power in signal `x`.
    The mean power of a Gaussian with `mu=0` and `sigma=1` is 1."""
    if variance is not None:
        variance = mean_square(variance)
    else:
        variance = 1.0

    return signal * numpy.sqrt(variance / mean_square(signal))


def heaviside(signal: numpy.ndarray) -> numpy.ndarray:
    """Heaviside.

    Returns the value 0 for `x < 0`, 1 for `x > 0`, and 1/2 for `x = 0`."""
    return 0.5 * (numpy.sign(signal) + 1)


if __name__ == "__main__":

    def asiudha() -> None:
        """
        :rtype: None
        """
        a = numpy.random.random(9) - 0.5
        print(
            normalise_signal_minmax(a),
            "\n",
            normalise_signal_max_abs(a),
            "\n",
            normalise_signal(a),
        )
        print(heaviside(a))

    asiudha()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"


# http://localhost:8097
# python -m visdom.server


def main():
    """
    Will start a visdom server
    """
    import visdom.server as server

    server.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15-02-2021
           """

from .annotate import *
from .color_utilities import *
from .figure_sessions import *
from .image_data import *
from .matplotlib_utilities import *
from .metric_overview_plot import *
from .quirks import *
from .signal_data import *
from .styles import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

from .boxes_2d import *
from .enums import *
from .sizes import *
from .torch_transforms import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from enum import Enum
from typing import Any

import cv2
import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/12/2019
           """

__all__ = ["cv2_resize", "InterpolationEnum"]


class InterpolationEnum(Enum):
    nearest = cv2.INTER_NEAREST  # a nearest-neighbor interpolation
    linear = cv2.INTER_LINEAR  # a bilinear interpolation (used by default)
    area = (
        cv2.INTER_AREA
    )  # resampling using pixel area relation. It may be a preferred method for image decimation, as it gives moire’-free results. But when the image is zoomed, it is similar to the INTER_NEAREST method.
    cubic = cv2.INTER_CUBIC  # a bicubic interpolation over 4×4 pixel neighborhood
    lanczos4 = cv2.INTER_LANCZOS4  # a Lanczos interpolation over 8×8 pixel neighborhood


def cv2_resize(
    x: Any,
    target_size: tuple,
    interpolation: InterpolationEnum = InterpolationEnum.linear,
) -> Any:
    """

    :param interpolation:
    :param x:
    :param target_size: proper (width, height) shape, no cv craziness
    :return:"""
    interpolation = InterpolationEnum(interpolation)
    if x.shape != target_size:
        x = cv2.resize(x, target_size[::-1], interpolation=interpolation.value)
    return x


if __name__ == "__main__":
    a = numpy.zeros((50, 50))
    a = cv2_resize(numpy.zeros((100, 100)), a.shape)
    print(a.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18-02-2021
           """

from typing import Iterable, List

import numpy
import pandas
import seaborn
from matplotlib import pyplot
from matplotlib.axes import Axes

from warg import Number

__all__ = [
    "despine_all",
    "set_y_log_scale",
    "exponential_moving_average",
]


def despine_all(ax: Axes = None) -> None:
    """

    :param ax:
    """
    if ax is None:
        ax = pyplot.gca()

    seaborn.despine(
        ax=ax, top=True, right=True, left=True, bottom=True, offset=None, trim=False
    )


def set_y_log_scale(ax: Axes = None) -> None:
    """

    :param ax:
    """
    if ax is None:
        ax = pyplot.gca()

    ax.set(yscale="log")


def exponential_moving_average(
    scalars: Iterable[Number], decay: float = 0.4
) -> List[Number]:
    """
    Like is usual in tensorboard visual rep just weight is inverse

    :param decay:
    :param scalars:
    :return:
    """
    if isinstance(scalars, numpy.ndarray):
        assert len(scalars.shape) <= 1, "only support one dimensional series"

    last = next(iter(scalars))
    smoothed = list()
    for new in scalars:
        # 1st-order IIR low-pass filter to attenuate the higher-
        # frequency components of the time-series.
        smoothed_point = last * decay + new * (1.0 - decay)
        smoothed.append(smoothed_point)
        last = smoothed_point

    return smoothed


if __name__ == "__main__":

    def stest_box_plot_props() -> None:
        """
        :rtype: None
        """
        props = {
            "boxprops": {"facecolor": "none", "edgecolor": "red"},
            "medianprops": {"color": "green"},
            "whiskerprops": {"color": "blue"},
            "capprops": {"color": "yellow"},
        }

        seaborn.boxplot(
            x="variable",
            y="value",
            data=pandas.DataFrame(
                [range(3), range(3)], columns=("variable", "value", "value2")
            ),
            showfliers=False,
            linewidth=0.75,
            **props
        )
        pyplot.show()

    stest_box_plot_props()



from torch.utils.data.sampler import BatchSampler


class LimitedBatchResampler(BatchSampler):
    """
    Wraps a BatchSampler, re-sampling from it until
    a specified number of iterations have been sampled"""

    def __init__(self, batch_sampler, num_iterations, start_iter: int = 0):
        self.batch_sampler = batch_sampler
        self.num_iterations = num_iterations
        self.start_iter = start_iter

    def __iter__(self):
        iteration = self.start_iter
        while iteration <= self.num_iterations:
            # if the underlying sampler has a set_epoch method, like
            # DistributedSampler, used for making each process see
            # a different split of the dataset, then set it
            if hasattr(self.batch_sampler.sampler, "set_epoch"):
                self.batch_sampler.sampler.set_epoch(iteration)
            for batch in self.batch_sampler:
                iteration += 1
                if iteration > self.num_iterations:
                    break
                yield batch

    def __len__(self):
        return self.num_iterations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-02-2021
           """

import timeit


def conversion_benchmark() -> None:
    """
    :rtype: None
    """
    # Numpy to matlab
    print(
        "\nCompare Numpy to matlab conversion strategy : (a bit long with several matlab.engine opening)"
    )
    setup_np2mat = (
        "import numpy\n"
        "import matlab\n"
        "from draugr.matlab_utilities import matlab_to_ndarray, ndarray_to_matlab\n"
        "import array\n"
        "np_a=numpy.random.uniform(size=(10000))*(.5+0.1236*1j) \n"
    )
    print(
        " > From matlab.double(np_a.tolist()) : "
        + str(
            timeit.timeit(
                "mat_a = matlab.double(np_a.real.tolist())",
                setup=setup_np2mat,
                number=100,
            )
        )
        + " s"
    )
    print(
        " > From ndarray_to_matlab [use pre alloc] : "
        + str(
            timeit.timeit(
                "mat_a = ndarray_to_matlab(np_a)", setup=setup_np2mat, number=100
            )
        )
        + " s"
    )

    # Matlab to numpy
    print("\nCompare matlab to numpy conversition strategy :")
    setup_tolist = (
        "import numpy\n"
        "import matlab\n"
        "from draugr.matlab_utilities import matlab_to_ndarray, ndarray_to_matlab\n"
        "eng = matlab.engine.start_matlab()\n"
        "mrd = eng.rand(matlab.int64([1,10000]),nargout=1)\n"
    )
    print(
        " > From numpy.array : "
        + str(
            timeit.timeit(
                "nprd = numpy.array(mrd,dtype = float) ", setup=setup_tolist, number=100
            )
        )
        + " s"
    )

    print(
        " > From numpy.asarray [use _data] : "
        + str(
            timeit.timeit(
                "nprd = numpy.asarray(mrd._data,dtype = float) ",
                setup=setup_tolist,
                number=100,
            )
        )
        + " s"
    )

    print(
        " > From matlab_to_ndarray [use _data buffer] : "
        + str(
            timeit.timeit(
                "nprd = matlab_to_ndarray(mrd) ", setup=setup_tolist, number=100
            )
        )
        + " s"
    )
    setup_tolist_cpx = (
        "import numpy\n"
        "import matlab\n"
        "from draugr.matlab_utilities import matlab_to_ndarray, ndarray_to_matlab\n"
        "eng = matlab.engine.start_matlab()\n"
        "mrd = eng.log(eng.linspace(-1.,1.,10000.))\n"
    )
    print(
        " > From matlab_to_ndarray [use _real _imag buffer complex] : "
        + str(
            timeit.timeit(
                "nprd = matlab_to_ndarray(mrd) ", setup=setup_tolist_cpx, number=100
            )
        )
        + " s"
    )


if __name__ == "__main__":
    conversion_benchmark()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02/03/2020
           """

__all__ = ["NoData", "IncompatiblePackageVersions"]

import types
from typing import Iterable, MutableMapping, Union


class NoData(Exception):
    """ """

    def __init__(self, msg: str = "No Data Available"):
        Exception.__init__(self, msg)


class IncompatiblePackageVersions(Exception):
    """ """

    def __init__(
        self, *packages: Iterable[Union[str, object]], **versions: MutableMapping
    ):
        str_o = ", "
        str_l = []

        for p in packages:
            if isinstance(p, str):
                s = f"({p},"
                if p in versions:
                    s += f"{versions[p]}"
                else:
                    s += f"NotSpecified"
                s += ")"
                str_l.append(s)
            elif isinstance(p, types.ModuleType):
                str_l.append(f"{p.__name__, p.__version__}")
        for vk, vv in versions.items():
            if vk not in packages:
                str_l.append(f"({vk},{vv})")

        Exception.__init__(self, f"Packages {str_o.join(str_l)} are not compatible")


if __name__ == "__main__":

    def main() -> None:
        """
        :rtype: None
        """
        raise IncompatiblePackageVersions(
            "numpy",
            "scipy",
            "Something",
            numpy="0.0.1",
            scipy="0.0.2",
            some_other_packge="0.5.0",
        )

    def main2() -> None:
        """
        :rtype: None
        """
        import numpy
        import scipy

        raise IncompatiblePackageVersions(numpy, scipy)

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-12-2020
           """

from typing import Sequence

import numpy

from apppath import ensure_existence
from draugr.numpy_utilities.datasets.splitting import SplitEnum
from draugr.numpy_utilities.signal_utilities.signal_statistics import root_mean_square
from draugr.numpy_utilities.signal_utilities.truncation import min_length_truncate_batch

LOG_MAGNITUDE_MULTIPLIER = 20
LOG_POWER_MULTIPLIER = 10

__all__ = ["mix_ratio"]


def mix_ratio(
    s1: Sequence,
    s2: Sequence,
    db_ratio: float = 0,
    *,
    log_multiplier: int = LOG_MAGNITUDE_MULTIPLIER,
) -> Sequence:
    """
    # Function to mix clean speech and noise at various SNR levels

    # Normalizing to -25 dB FS

    :param s1:
    :param s2:
    :param db_ratio:
    :param log_multiplier:
    :return:"""
    s1, s2 = min_length_truncate_batch((s1, s2))
    s1_max = numpy.abs(s1).max()
    mix = (s1 / s1_max) + (
        (s2 / numpy.abs(s2).max()) * (root_mean_square(s1) / root_mean_square(s2))
    ) / (10 ** (db_ratio / log_multiplier))
    return (mix / numpy.abs(mix).max()) * s1_max


if __name__ == "__main__":

    def asad() -> None:
        """
        :rtype: None
        """
        from neodroidaudition.data.recognition.libri_speech import LibriSpeech
        from neodroidaudition.noise_generation.gaussian_noise import white_noise

        from pathlib import Path

        libri_speech = LibriSpeech(
            path=Path.home() / "Data" / "Audio" / "Speech" / "LibriSpeech"
        )
        files, sr = zip(*[(v[0].numpy(), v[1]) for _, v in zip(range(1), libri_speech)])
        assert all([sr[0] == s for s in sr[1:]])

        mixed = mix_ratio(files[0], white_noise(files[0].shape[-1]))
        print(mixed)

    def asadsa() -> None:
        """
        :rtype: None
        """
        from draugr.torch_utilities import to_tensor
        from neodroidaudition.data.recognition.libri_speech import LibriSpeech
        from neodroidaudition.noise_generation.gaussian_noise import white_noise
        import torchaudio

        from pathlib import Path

        libri_speech = LibriSpeech(
            path=Path.home() / "Data" / "Audio" / "Speech" / "LibriSpeech",
            split=SplitEnum.testing,
        )
        files, sr = zip(
            *[(v[0].numpy(), v[1]) for _, v in zip(range(20), libri_speech)]
        )
        assert all([sr[0] == s for s in sr[1:]])

        mix = files[0]
        for file in files[1:]:
            mix = mix_ratio(mix, file, 0)

        torchaudio.save(
            str(ensure_existence(Path.cwd() / "exclude") / f"mixed_even_babble.wav"),
            to_tensor(mix),
            int(sr[0]),
        )

        for ratio in range(-20, 20 + 1, 5):
            torchaudio.save(
                str(ensure_existence(Path.cwd() / "exclude") / f"mixed_{ratio}.wav"),
                to_tensor(mix_ratio(files[0], files[-1], ratio)),
                int(sr[0]),
            )

    def asadsa2() -> None:
        """
        :rtype: None
        """
        from draugr.torch_utilities import to_tensor
        from neodroidaudition.data.recognition.libri_speech import LibriSpeech
        from neodroidaudition.noise_generation.gaussian_noise import white_noise
        import torchaudio
        from pathlib import Path

        libri_speech = LibriSpeech(
            path=Path.home() / "Data" / "Audio" / "Speech" / "LibriSpeech"
        )
        files, sr = zip(*[(v[0].numpy(), v[1]) for _, v in zip(range(1), libri_speech)])
        assert all([sr[0] == s for s in sr[1:]])

        normed = files[0]
        mixed = mix_ratio(normed, normed, 0)
        mixed2 = mix_ratio(mixed, mixed, 0)
        print(normed, mixed)
        print(mixed2, mixed)
        print(root_mean_square(normed))
        print(root_mean_square(mixed))
        print(root_mean_square(mixed2))
        assert numpy.allclose(normed, mixed)
        assert numpy.allclose(mixed2, mixed)
        torchaudio.save(
            str(ensure_existence(Path.cwd() / "exclude") / "mixed_same.wav"),
            to_tensor(mixed),
            int(sr[0]),
        )

    # asad()
    asadsa()
    # asadsa2()



from typing import Tuple

import numpy
import torch

__all__ = [
    "CV2ToImage",
    "CV2ToTensor",
]


class CV2ToImage(object):
    def __call__(
        self,
        tensor: torch.Tensor,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        return (
            tensor.cpu().numpy().astype(numpy.float32).transpose((1, 2, 0)),
            boxes,
            labels,
        )


class CV2ToTensor(object):
    def __call__(
        self,
        cvimage: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        return (
            torch.from_numpy(cvimage.astype(numpy.float32)).permute(2, 0, 1),
            boxes,
            labels,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-02-2021
           """

import numpy
import torch
from torch.autograd import Function

__all__ = ["GuidedBackPropReLUModel", "GuidedBackPropReLU"]


class GuidedBackPropReLU(Function):
    @staticmethod
    def forward(self, input_img):
        """

        :param self:
        :param input_img:
        :return:
        """
        positive_mask = (input_img > 0).type_as(input_img)
        output = torch.addcmul(
            torch.zeros(input_img.size()).type_as(input_img), input_img, positive_mask
        )
        self.save_for_backward(input_img, output)
        return output

    @staticmethod
    def backward(self, grad_output):
        """

        :param self:
        :param grad_output:
        :return:
        """
        input_img, output = self.saved_tensors

        positive_mask_1 = (input_img > 0).type_as(grad_output)
        positive_mask_2 = (grad_output > 0).type_as(grad_output)
        grad_input = torch.addcmul(
            torch.zeros(input_img.size()).type_as(input_img),
            torch.addcmul(
                torch.zeros(input_img.size()).type_as(input_img),
                grad_output,
                positive_mask_1,
            ),
            positive_mask_2,
        )
        return grad_input


class GuidedBackPropReLUModel:
    """ """

    def __init__(self, model, use_cuda):
        self._model = model
        self._model.eval()
        self._use_cuda = use_cuda
        if self._use_cuda:
            self._model = self._model.cuda()

        def recursive_relu_apply(module_top):
            """

            :param module_top:
            """
            for idx, module in module_top._modules.items():
                recursive_relu_apply(module)
                if module.__class__.__name__ == "ReLU":
                    module_top._modules[idx] = GuidedBackPropReLU.apply

        recursive_relu_apply(self._model)  # replace ReLU with GuidedBackpropReLU

    def forward(self, input_img):
        """

        :param input_img:
        :return:
        """
        return self._model(input_img)

    def __call__(self, input_img, target_category=None):
        if self._use_cuda:
            input_img = input_img.cuda()

        input_img = input_img.requires_grad_(True)

        output = self.forward(input_img)

        if target_category is None:
            target_category = numpy.argmax(output.cpu().data.numpy())

        one_hot = numpy.zeros((1, output.size()[-1]), dtype=numpy.float32)
        one_hot[0][target_category] = 1
        one_hot = torch.from_numpy(one_hot).requires_grad_(True)
        if self._use_cuda:
            one_hot = one_hot.cuda()

        one_hot = torch.sum(one_hot * output)
        one_hot.backward(retain_graph=True)

        output = input_img.grad.cpu().data.numpy()
        output = output[0, :, :, :]

        return output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 24/01/2022
           """

# TODO: Make instanced Terminal plotter for replotting console contents with new data



from enum import Enum

import cv2

__all__ = ["ColorConversionEnum"]


class ColorConversionEnum(Enum):
    """
    the color conversion codes
    """

    bgr2bgra = cv2.COLOR_BGR2BGRA  # add alpha channel to RGB or BGR image
    rgb2rgba = cv2.COLOR_RGB2RGBA

    bgra2bgr = cv2.COLOR_BGRA2BGR  # remove alpha channel from RGB or BGR image
    rgba2rgb = cv2.COLOR_RGBA2RGB

    # ---
    bgr2rgba = (
        cv2.COLOR_BGR2RGBA
    )  # convert between RGB and BGR color spaces (with or without alpha channel)
    rgb2bgra = cv2.COLOR_RGB2BGRA

    rgba2bgr = cv2.COLOR_RGBA2BGR
    bgra2rgb = cv2.COLOR_BGRA2RGB

    bgr2rgb = cv2.COLOR_BGR2RGB
    rgb2bgr = cv2.COLOR_RGB2BGR

    bgra2rgba = cv2.COLOR_BGRA2RGBA
    rgba2bgra = cv2.COLOR_RGBA2BGRA

    # ---
    bgr2gray = (
        cv2.COLOR_BGR2GRAY
    )  # convert between RGB/BGR and grayscale, color conversions
    rgb2gray = cv2.COLOR_RGB2GRAY
    bgra2gray = cv2.COLOR_BGRA2GRAY
    rgba2gray = cv2.COLOR_RGBA2GRAY

    gray2bgr = cv2.COLOR_GRAY2BGR  # grayscale, color conversions
    gray2rgb = cv2.COLOR_GRAY2RGB
    gray2bgra = cv2.COLOR_GRAY2BGRA
    gray2rgba = cv2.COLOR_GRAY2RGBA
    # ---

    bgr2bgr565 = (
        cv2.COLOR_BGR2BGR565
    )  # convert between RGB/BGR and BGR565 (16-bit images)
    rgb2bgr565 = cv2.COLOR_RGB2BGR565
    bgr5652bgr = cv2.COLOR_BGR5652BGR
    bgr5652rgb = cv2.COLOR_BGR5652RGB
    bgra2bgr565 = cv2.COLOR_BGRA2BGR565
    rgba2bgr565 = cv2.COLOR_RGBA2BGR565
    bgr5652bgra = cv2.COLOR_BGR5652BGRA
    bgr5652rgba = cv2.COLOR_BGR5652RGBA
    # -----

    gray2bgr565 = (
        cv2.COLOR_GRAY2BGR565
    )  # convert between grayscale to BGR565 (16-bit images)
    bgr5652gray = cv2.COLOR_BGR5652GRAY

    # --- ---
    bgr2bgr555 = (
        cv2.COLOR_BGR2BGR555
    )  # convert between RGB/BGR and BGR555 (16-bit images)
    rgb2bgr555 = cv2.COLOR_RGB2BGR555
    bgr5552bgr = cv2.COLOR_BGR5552BGR
    bgr5552rgb = cv2.COLOR_BGR5552RGB
    bgra2bgr555 = cv2.COLOR_BGRA2BGR555
    rgba2bgr555 = cv2.COLOR_RGBA2BGR555
    bgr5552bgra = cv2.COLOR_BGR5552BGRA
    bgr5552rgba = cv2.COLOR_BGR5552RGBA

    # -- ---
    gray2bgr555 = (
        cv2.COLOR_GRAY2BGR555
    )  # convert between grayscale and BGR555 (16-bit images)

    bgr5552gray = cv2.COLOR_BGR5552GRAY

    # -- ---
    bgr2xyz = cv2.COLOR_BGR2XYZ  # convert RGB/BGR to CIE XYZ, color conversions
    rgb2xyz = cv2.COLOR_RGB2XYZ
    xyz2bgr = cv2.COLOR_XYZ2BGR
    xyz2rgb = cv2.COLOR_XYZ2RGB

    # -- ---
    bgr2ycrcb = (
        cv2.COLOR_BGR2YCrCb
    )  # convert RGB/BGR to luma-chroma (aka YCC), color conversions
    rgb2ycrcb = cv2.COLOR_RGB2YCrCb
    ycrcb2bgr = cv2.COLOR_YCrCb2BGR
    ycrcb2rgb = cv2.COLOR_YCrCb2RGB

    # -- ---
    bgr2hsv = (
        cv2.COLOR_BGR2HSV
    )  # convert RGB/BGR to HSV (hue saturation value) with H range 0..180 if 8 bit image, color conversions
    rgb2hsv = cv2.COLOR_RGB2HSV

    # 37 ---
    bgr2lab = cv2.COLOR_BGR2Lab  # convert RGB/BGR to CIE Lab, color conversions
    rgb2lab = cv2.COLOR_RGB2Lab

    # 40 --
    bgr2luv = cv2.COLOR_BGR2Luv  # convert RGB/BGR to CIE Luv, color conversions
    rgb2luv = cv2.COLOR_RGB2Luv

    # -- ---
    bgr2hls = (
        cv2.COLOR_BGR2HLS
    )  # convert RGB/BGR to HLS (hue lightness saturation) with H range 0..180 if 8 bit image, color conversions
    rgb2hls = cv2.COLOR_RGB2HLS

    # --- ---
    hsv2bgr = (
        cv2.COLOR_HSV2BGR
    )  # backward conversions HSV to RGB/BGR with H range 0..180 if 8 bit image
    hsv2rgb = cv2.COLOR_HSV2RGB
    lab2bgr = cv2.COLOR_Lab2BGR
    lab2rgb = cv2.COLOR_Lab2RGB
    luv2bgr = cv2.COLOR_Luv2BGR
    luv2rgb = cv2.COLOR_Luv2RGB

    # -- ---
    hls2bgr = (
        cv2.COLOR_HLS2BGR
    )  # backward conversions HLS to RGB/BGR with H range 0..180 if 8 bit image
    hls2rgb = cv2.COLOR_HLS2RGB

    # -- ---
    bgr2hsv_full = (
        cv2.COLOR_BGR2HSV_FULL
    )  # convert RGB/BGR to HSV (hue saturation value) with H range 0..255 if 8 bit image, color conversions
    rgb2hsv_full = cv2.COLOR_RGB2HSV_FULL

    # -- ---
    bgr2hls_full = (
        cv2.COLOR_BGR2HLS_FULL
    )  # convert RGB/BGR to HLS (hue lightness saturation) with H range 0..255 if 8 bit image, color conversions

    rgb2hls_full = cv2.COLOR_RGB2HLS_FULL

    # --- ---
    hsv2bgr_full = (
        cv2.COLOR_HSV2BGR_FULL
    )  # backward conversions HSV to RGB/BGR with H range 0..255 if 8 bit image

    hsv2rgb_full = cv2.COLOR_HSV2RGB_FULL

    # --- ---
    hls2bgr_full = (
        cv2.COLOR_HLS2BGR_FULL
    )  # backward conversions HLS to RGB/BGR with H range 0..255 if 8 bit image

    hls2rgb_full = cv2.COLOR_HLS2RGB_FULL

    lbgr2lab = cv2.COLOR_LBGR2Lab

    lrgb2lab = cv2.COLOR_LRGB2Lab

    lbgr2luv = cv2.COLOR_LBGR2Luv

    lrgb2luv = cv2.COLOR_LRGB2Luv

    lab2lbgr = cv2.COLOR_Lab2LBGR

    lab2lrgb = cv2.COLOR_Lab2LRGB

    luv2lbgr = cv2.COLOR_Luv2LBGR

    luv2lrgb = cv2.COLOR_Luv2LRGB

    # -- ---
    bgr2yuv = cv2.COLOR_BGR2YUV  # convert between RGB/BGR and YUV

    rgb2yuv = cv2.COLOR_RGB2YUV

    yuv2bgr = cv2.COLOR_YUV2BGR

    yuv2rgb = cv2.COLOR_YUV2RGB

    # ------
    yuv2rgb_nv12 = cv2.COLOR_YUV2RGB_NV12  # YUV 4:2:0 family to RGB.

    yuv2bgr_nv12 = cv2.COLOR_YUV2BGR_NV12

    yuv2rgb_nv21 = cv2.COLOR_YUV2RGB_NV21

    yuv2bgr_nv21 = cv2.COLOR_YUV2BGR_NV21

    yuv420sp2rgb = cv2.COLOR_YUV420sp2RGB

    yuv420sp2bgr = cv2.COLOR_YUV420sp2BGR

    yuv2rgba_nv12 = cv2.COLOR_YUV2RGBA_NV12

    yuv2bgra_nv12 = cv2.COLOR_YUV2BGRA_NV12

    yuv2rgba_nv21 = cv2.COLOR_YUV2RGBA_NV21

    yuv2bgra_nv21 = cv2.COLOR_YUV2BGRA_NV21

    yuv420sp2rgba = cv2.COLOR_YUV420sp2RGBA

    yuv420sp2bgra = cv2.COLOR_YUV420sp2BGRA

    yuv2rgb_yv12 = cv2.COLOR_YUV2RGB_YV12

    yuv2bgr_yv12 = cv2.COLOR_YUV2BGR_YV12

    yuv2rgb_iyuv = cv2.COLOR_YUV2RGB_IYUV

    yuv2bgr_iyuv = cv2.COLOR_YUV2BGR_IYUV

    yuv2rgb_i420 = cv2.COLOR_YUV2RGB_I420

    yuv2bgr_i420 = cv2.COLOR_YUV2BGR_I420

    yuv420p2rgb = cv2.COLOR_YUV420p2RGB

    yuv420p2bgr = cv2.COLOR_YUV420p2BGR

    yuv2rgba_yv12 = cv2.COLOR_YUV2RGBA_YV12

    yuv2bgra_yv12 = cv2.COLOR_YUV2BGRA_YV12

    yuv2rgba_iyuv = cv2.COLOR_YUV2RGBA_IYUV

    yuv2bgra_iyuv = cv2.COLOR_YUV2BGRA_IYUV

    yuv2rgba_i420 = cv2.COLOR_YUV2RGBA_I420

    yuv2bgra_i420 = cv2.COLOR_YUV2BGRA_I420

    yuv420p2rgba = cv2.COLOR_YUV420p2RGBA

    yuv420p2bgra = cv2.COLOR_YUV420p2BGRA

    yuv2gray_420 = cv2.COLOR_YUV2GRAY_420

    yuv2gray_nv21 = cv2.COLOR_YUV2GRAY_NV21

    yuv2gray_nv12 = cv2.COLOR_YUV2GRAY_NV12

    yuv2gray_yv12 = cv2.COLOR_YUV2GRAY_YV12

    yuv2gray_iyuv = cv2.COLOR_YUV2GRAY_IYUV

    yuv2gray_i420 = cv2.COLOR_YUV2GRAY_I420

    yuv420sp2gray = cv2.COLOR_YUV420sp2GRAY

    yuv420p2gray = cv2.COLOR_YUV420p2GRAY

    # -- ---
    yuv2rgb_uyvy = cv2.COLOR_YUV2RGB_UYVY  # YUV 4:2:2 family to RGB.

    yuv2bgr_uyvy = cv2.COLOR_YUV2BGR_UYVY

    yuv2rgb_y422 = cv2.COLOR_YUV2RGB_Y422

    yuv2bgr_y422 = cv2.COLOR_YUV2BGR_Y422

    yuv2rgb_uynv = cv2.COLOR_YUV2RGB_UYNV

    yuv2bgr_uynv = cv2.COLOR_YUV2BGR_UYNV

    yuv2rgba_uyvy = cv2.COLOR_YUV2RGBA_UYVY

    yuv2bgra_uyvy = cv2.COLOR_YUV2BGRA_UYVY

    yuv2rgba_y422 = cv2.COLOR_YUV2RGBA_Y422

    yuv2bgra_y422 = cv2.COLOR_YUV2BGRA_Y422

    yuv2rgba_uynv = cv2.COLOR_YUV2RGBA_UYNV

    yuv2bgra_uynv = cv2.COLOR_YUV2BGRA_UYNV

    yuv2rgb_yuy2 = cv2.COLOR_YUV2RGB_YUY2

    yuv2bgr_yuy2 = cv2.COLOR_YUV2BGR_YUY2

    yuv2rgb_yvyu = cv2.COLOR_YUV2RGB_YVYU

    yuv2bgr_yvyu = cv2.COLOR_YUV2BGR_YVYU

    yuv2rgb_yuyv = cv2.COLOR_YUV2RGB_YUYV

    yuv2bgr_yuyv = cv2.COLOR_YUV2BGR_YUYV

    yuv2rgb_yunv = cv2.COLOR_YUV2RGB_YUNV

    yuv2bgr_yunv = cv2.COLOR_YUV2BGR_YUNV

    yuv2rgba_yuy2 = cv2.COLOR_YUV2RGBA_YUY2

    yuv2bgra_yuy2 = cv2.COLOR_YUV2BGRA_YUY2

    yuv2rgba_yvyu = cv2.COLOR_YUV2RGBA_YVYU

    yuv2bgra_yvyu = cv2.COLOR_YUV2BGRA_YVYU

    yuv2rgba_yuyv = cv2.COLOR_YUV2RGBA_YUYV

    yuv2bgra_yuyv = cv2.COLOR_YUV2BGRA_YUYV

    yuv2rgba_yunv = cv2.COLOR_YUV2RGBA_YUNV

    yuv2bgra_yunv = cv2.COLOR_YUV2BGRA_YUNV

    yuv2gray_uyvy = cv2.COLOR_YUV2GRAY_UYVY

    yuv2gray_yuy2 = cv2.COLOR_YUV2GRAY_YUY2

    yuv2gray_y422 = cv2.COLOR_YUV2GRAY_Y422
    yuv2gray_uynv = cv2.COLOR_YUV2GRAY_UYNV

    yuv2gray_yvyu = cv2.COLOR_YUV2GRAY_YVYU

    yuv2gray_yuyv = cv2.COLOR_YUV2GRAY_YUYV

    yuv2gray_yunv = cv2.COLOR_YUV2GRAY_YUNV

    # -----
    rgba2mrgba = cv2.COLOR_RGBA2mRGBA  # alpha premultiplication

    mrgba2rgba = cv2.COLOR_mRGBA2RGBA

    # -- ---
    rgb2yuv_i420 = cv2.COLOR_RGB2YUV_I420  # RGB to YUV 4:2:0 family.

    bgr2yuv_i420 = cv2.COLOR_BGR2YUV_I420

    rgb2yuv_iyuv = cv2.COLOR_RGB2YUV_IYUV

    bgr2yuv_iyuv = cv2.COLOR_BGR2YUV_IYUV

    rgba2yuv_i420 = cv2.COLOR_RGBA2YUV_I420

    bgra2yuv_i420 = cv2.COLOR_BGRA2YUV_I420

    rgba2yuv_iyuv = cv2.COLOR_RGBA2YUV_IYUV

    bgra2yuv_iyuv = cv2.COLOR_BGRA2YUV_IYUV

    rgb2yuv_yv12 = cv2.COLOR_RGB2YUV_YV12

    bgr2yuv_yv12 = cv2.COLOR_BGR2YUV_YV12

    rgba2yuv_yv12 = cv2.COLOR_RGBA2YUV_YV12

    bgra2yuv_yv12 = cv2.COLOR_BGRA2YUV_YV12

    # --- ---
    bayerbg2bgr = (
        cv2.COLOR_BayerBG2BGR
    )  # Demosaicing, see color conversions for additional information. #equivalent to RGGB Bayer pattern

    bayergb2bgr = cv2.COLOR_BayerGB2BGR  # equivalent to GRBG Bayer pattern

    bayerrg2bgr = cv2.COLOR_BayerRG2BGR  # equivalent to BGGR Bayer pattern

    bayergr2bgr = cv2.COLOR_BayerGR2BGR  # equivalent to GBRG Bayer pattern

    bayerrggb2bgr = cv2.COLOR_BayerRGGB2BGR

    bayergrbg2bgr = cv2.COLOR_BayerGRBG2BGR

    bayerbggr2bgr = cv2.COLOR_BayerBGGR2BGR

    bayergbrg2bgr = cv2.COLOR_BayerGBRG2BGR

    bayerrggb2rgb = cv2.COLOR_BayerRGGB2RGB

    bayergrbg2rgb = cv2.COLOR_BayerGRBG2RGB

    bayerbggr2rgb = cv2.COLOR_BayerBGGR2RGB

    bayergbrg2rgb = cv2.COLOR_BayerGBRG2RGB

    bayerbg2rgb = cv2.COLOR_BayerBG2RGB  # equivalent to RGGB Bayer pattern

    bayergb2rgb = cv2.COLOR_BayerGB2RGB  # equivalent to GRBG Bayer pattern

    bayerrg2rgb = cv2.COLOR_BayerRG2RGB  # equivalent to BGGR Bayer pattern

    bayergr2rgb = cv2.COLOR_BayerGR2RGB  # equivalent to GBRG Bayer pattern

    bayerbg2gray = cv2.COLOR_BayerBG2GRAY  # equivalent to RGGB Bayer pattern

    bayergb2gray = cv2.COLOR_BayerGB2GRAY  # equivalent to GRBG Bayer pattern

    bayerrg2gray = cv2.COLOR_BayerRG2GRAY  # equivalent to BGGR Bayer pattern

    bayergr2gray = cv2.COLOR_BayerGR2GRAY  # equivalent to GBRG Bayer pattern

    bayerrggb2gray = cv2.COLOR_BayerRGGB2GRAY

    bayergrbg2gray = cv2.COLOR_BayerGRBG2GRAY

    bayerbggr2gray = cv2.COLOR_BayerBGGR2GRAY

    bayergbrg2gray = cv2.COLOR_BayerGBRG2GRAY

    # --- ---
    bayerbg2bgr_vng = (
        cv2.COLOR_BayerBG2BGR_VNG
    )  # Demosaicing using Variable Number of Gradients. #equivalent to RGGB Bayer pattern

    bayergb2bgr_vng = cv2.COLOR_BayerGB2BGR_VNG  # equivalent to GRBG Bayer pattern

    bayerrg2bgr_vng = cv2.COLOR_BayerRG2BGR_VNG  # equivalent to BGGR Bayer pattern

    bayergr2bgr_vng = cv2.COLOR_BayerGR2BGR_VNG  # equivalent to GBRG Bayer pattern

    bayerrggb2bgr_vng = cv2.COLOR_BayerRGGB2BGR_VNG

    bayergrbg2bgr_vng = cv2.COLOR_BayerGRBG2BGR_VNG

    bayerbggr2bgr_vng = cv2.COLOR_BayerBGGR2BGR_VNG

    bayergbrg2bgr_vng = cv2.COLOR_BayerGBRG2BGR_VNG

    bayerrggb2rgb_vng = cv2.COLOR_BayerRGGB2RGB_VNG

    bayergrbg2rgb_vng = cv2.COLOR_BayerGRBG2RGB_VNG

    bayerbggr2rgb_vng = cv2.COLOR_BayerBGGR2RGB_VNG

    bayergbrg2rgb_vng = cv2.COLOR_BayerGBRG2RGB_VNG

    bayerbg2rgb_vng = cv2.COLOR_BayerBG2RGB_VNG  # equivalent to RGGB Bayer pattern

    bayergb2rgb_vng = cv2.COLOR_BayerGB2RGB_VNG  # equivalent to GRBG Bayer pattern

    bayerrg2rgb_vng = cv2.COLOR_BayerRG2RGB_VNG  # equivalent to BGGR Bayer pattern

    bayergr2rgb_vng = cv2.COLOR_BayerGR2RGB_VNG  # equivalent to GBRG Bayer pattern

    bayerbg2bgr_ea = (
        cv2.COLOR_BayerBG2BGR_EA
    )  # Edge-Aware Demosaicing. #equivalent to RGGB Bayer pattern

    bayergb2bgr_ea = cv2.COLOR_BayerGB2BGR_EA  # equivalent to GRBG Bayer pattern

    bayerrg2bgr_ea = cv2.COLOR_BayerRG2BGR_EA  # equivalent to BGGR Bayer pattern

    bayergr2bgr_ea = cv2.COLOR_BayerGR2BGR_EA  # equivalent to GBRG Bayer pattern

    bayerrggb2bgr_ea = cv2.COLOR_BayerRGGB2BGR_EA

    bayergrbg2bgr_ea = cv2.COLOR_BayerGRBG2BGR_EA

    bayerbggr2bgr_ea = cv2.COLOR_BayerBGGR2BGR_EA

    bayergbrg2bgr_ea = cv2.COLOR_BayerGBRG2BGR_EA

    bayerrggb2rgb_ea = cv2.COLOR_BayerRGGB2RGB_EA

    bayergrbg2rgb_ea = cv2.COLOR_BayerGRBG2RGB_EA

    bayerbggr2rgb_ea = cv2.COLOR_BayerBGGR2RGB_EA

    bayergbrg2rgb_ea = cv2.COLOR_BayerGBRG2RGB_EA

    bayerbg2rgb_ea = cv2.COLOR_BayerBG2RGB_EA  # equivalent to RGGB Bayer pattern

    bayergb2rgb_ea = cv2.COLOR_BayerGB2RGB_EA  # equivalent to GRBG Bayer pattern

    bayerrg2rgb_ea = cv2.COLOR_BayerRG2RGB_EA  # equivalent to BGGR Bayer pattern

    bayergr2rgb_ea = cv2.COLOR_BayerGR2RGB_EA  # equivalent to GBRG Bayer pattern

    # --- ---
    bayerbg2bgra = (
        cv2.COLOR_BayerBG2BGRA
    )  # Demosaicing with alpha channel. #equivalent to RGGB Bayer pattern

    bayergb2bgra = cv2.COLOR_BayerGB2BGRA  # equivalent to GRBG Bayer pattern

    bayerrg2bgra = cv2.COLOR_BayerRG2BGRA  # equivalent to BGGR Bayer pattern

    bayergr2bgra = cv2.COLOR_BayerGR2BGRA  # equivalent to GBRG Bayer pattern

    bayerrggb2bgra = cv2.COLOR_BayerRGGB2BGRA

    bayergrbg2bgra = cv2.COLOR_BayerGRBG2BGRA

    bayerbggr2bgra = cv2.COLOR_BayerBGGR2BGRA

    bayergbrg2bgra = cv2.COLOR_BayerGBRG2BGRA

    bayerrggb2rgba = cv2.COLOR_BayerRGGB2RGBA

    bayergrbg2rgba = cv2.COLOR_BayerGRBG2RGBA

    bayerbggr2rgba = cv2.COLOR_BayerBGGR2RGBA

    bayergbrg2rgba = cv2.COLOR_BayerGBRG2RGBA

    bayerbg2rgba = cv2.COLOR_BayerBG2RGBA  # equivalent to RGGB Bayer pattern
    bayergb2rgba = cv2.COLOR_BayerGB2RGBA  # equivalent to GRBG Bayer pattern
    bayerrg2rgba = cv2.COLOR_BayerRG2RGBA  # equivalent to BGGR Bayer pattern
    bayergr2rgba = cv2.COLOR_BayerGR2RGBA  # equivalent to GBRG Bayer pattern

    # -----
    colorcvt_max = cv2.COLOR_COLORCVT_MAX


if __name__ == "__main__":
    print(ColorConversionEnum.rgb2rgba, ColorConversionEnum.rgb2rgba.value)



import os
from enum import Enum
from typing import Iterable, Union

__all__ = ["DeviceOrderEnum", "set_cuda_device_order", "set_cuda_visible_devices"]


class DeviceOrderEnum(Enum):
    fastest_first = "FASTEST_FIRST"
    pci_bus_id = "PCI_BUS_ID"


def set_cuda_device_order(order: DeviceOrderEnum = DeviceOrderEnum.pci_bus_id) -> None:
    os.environ["CUDA_DEVICE_ORDER"] = DeviceOrderEnum(order).value


def set_cuda_visible_devices(devices: Union[str, int, Iterable[int]]) -> None:
    if isinstance(devices, int):
        devices = str(devices)
    elif isinstance(devices, Iterable):
        devices = ",".join(str(d) for d in devices)
    if devices is None:  # TODO: Nix specific, choose the least utilised device
        devices = "$(nvidia-smi --query-gpu=memory.free,index --format=csv,nounits,noheader | sort -nr | head -1 | awk '{ print $NF }')"
    os.environ["CUDA_VISIBLE_DEVICES"] = devices



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-02-2021
           """

from matplotlib import cycler, pyplot, rcParams
from matplotlib.axes import Axes

from draugr.visualisation.matplotlib_utilities.styles.cyclers import simple_hatch_cycler

__all__ = [
    "fix_edge_gridlines",
    "auto_post_print_dpi",
    "auto_post_hatch",
    "scatter_auto_mark",
]

from warg import Number, drop_unused_kws, passes_kws_to
import numpy


def fix_edge_gridlines(
    ax: Axes = None,
) -> None:  # TODO: make a wrapper version of the function
    """
    Fixes gridlines when using zero margins round_number number ticks

    :param ax:
    :return:"""
    if ax is None:
        ax = pyplot.gca()
    ax.xaxis.get_gridlines()[-1].set_clip_on(False)  # last gridline
    ax.yaxis.get_gridlines()[0].set_clip_on(False)  # first gridline


def auto_post_print_dpi(scalar: Number = 5) -> None:
    """
    auto scale dpi of lines for print

    :param scalar:
    :return:"""
    dpi = rcParams["figure.dpi"]
    line_width = 1.0 / dpi * scalar
    rcParams["hatch.linewidth"] = line_width * 2
    rcParams["grid.linewidth"] = line_width


def auto_post_hatch(
    ax: Axes = None, hatch_cycler: cycler = simple_hatch_cycler
) -> None:
    """
    Auto hatches patch-types because matplotlib's prop-cycler does not iterate hatch_props

    :param ax:
    :param hatch_cycler:
    :return:"""
    if ax is None:
        ax = pyplot.gca()
    for p, d in zip(ax.patches, hatch_cycler):
        p.set_hatch(next(iter(d.values())))


@drop_unused_kws
@passes_kws_to(pyplot.scatter)
def scatter_auto_mark(x, y, c, ax=None, m=("|", "_"), fillstyle="none", **kw):
    """
    TODO:Quick hack, can be generalised further
    :param x:
    :param y:
    :param c:
    :param ax:
    :param m:
    :param fillstyle:
    :param kw:
    :return:
    """
    import matplotlib.markers as mmarkers

    if not ax:
        ax = pyplot.gca()
    sc = ax.scatter(x, y, c=c, **kw)
    if m is not None and len(m) == len(x):
        paths = []
        for marker in m:
            if isinstance(marker, mmarkers.MarkerStyle):
                marker_obj = marker
            else:
                marker_obj = mmarkers.MarkerStyle(marker, fillstyle=fillstyle)
            path = marker_obj.get_path().transformed(marker_obj.get_transform())
            paths.append(path)
        sc.set_paths(paths)
    elif (
        c is not None and isinstance(c[0], (int, numpy.ndarray)) and len(c) == len(x)
    ):  # TODO: HANDLE numpy ndarray
        # better
        paths = []
        for c_ in c:
            if isinstance(c_, numpy.ndarray):
                c_ = c_.item()
            if isinstance(m[c_], mmarkers.MarkerStyle):
                marker_obj = m[c_]
            else:
                marker_obj = mmarkers.MarkerStyle(m[c_], fillstyle=fillstyle)
            paths.append(marker_obj.get_path().transformed(marker_obj.get_transform()))
        sc.set_paths(paths)
    else:
        pass
        # raise NotImplemented
    return sc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-12-2020
           """

import time
from enum import Enum
from itertools import tee
from typing import Any, Generator, Iterable

import tqdm
from sorcery import assigned_names

from notus.notification import JobNotificationSession
from warg import drop_unused_kws, passes_kws_to, empty_str

__all__ = ["progress_bar"]


class TimestampModeEnum(Enum):
    none, prefix, postfix = assigned_names()


@drop_unused_kws
@passes_kws_to(tqdm.tqdm)
def progress_bar(
    iterable: Iterable,
    description: str = None,
    *,
    leave: bool = False,
    notifications: bool = False,
    total: int = None,
    auto_total_generator: bool = False,
    auto_describe_iterator: bool = True,  # DOES NOT WORK IS THIS FUNCTION IS ALIAS does not match!
    alias="progress_bar",
    disable: bool = False,
    verbose: bool = False,
    timestamp_mode: TimestampModeEnum = TimestampModeEnum.none,
    **kwargs,
) -> Any:
    """
    hint:
    use next first and then later use send instead of next to set a new tqdm description if desired.
    """
    if not disable:
        if description is None and auto_describe_iterator:
            from warg import get_first_arg_name

            description = get_first_arg_name(alias, verbose=verbose)

        if total is None and isinstance(iterable, Generator) and auto_total_generator:
            iterable, ic = tee(iterable, 2)
            total = len(list(ic))
            if total == 0:
                print(f"WARNING zero length iterable - {description}:{iterable}")

        generator = tqdm.tqdm(
            iterable,
            description,
            leave=leave,
            total=total,
            disable=disable,  # redundant
            **kwargs,
        )
        prefix = empty_str
        postfix = empty_str
        update_timestamp = True
        if timestamp_mode == TimestampModeEnum.prefix:
            prefix = time.time
        elif timestamp_mode == TimestampModeEnum.postfix:
            postfix = time.time
        else:
            update_timestamp = False

        if notifications:
            with JobNotificationSession(description):
                for val in generator:
                    a = yield val
                    if update_timestamp or a:
                        if a:
                            description = str(a)
                        generator.set_description(
                            " ".join(
                                (str(prefix()), description, str(postfix()))
                            ).strip()
                        )
            return
        for val in generator:
            a = yield val
            if update_timestamp or a:
                if a:
                    description = str(a)
                generator.set_description(
                    " ".join((str(prefix()), description, str(postfix()))).strip()
                )
    else:
        yield from iterable


if __name__ == "__main__":

    def dsad3123() -> None:
        """
        :rtype: None
        """
        from time import sleep

        for a in progress_bar([2.13, 8921.9123, 923], notifications=False):
            sleep(1)

    def asd21sa() -> None:
        """
        :rtype: None
        """
        from time import sleep

        pb = progress_bar  # Aliased! does not find description at the moment

        for a in pb([2.13, 8921.9123, 923], notifications=False):
            sleep(1)

    def dict_items() -> None:
        """
        :rtype: None
        """
        from time import sleep

        class exp_v:
            Test_Sets = {v: v for v in range(9)}

        for a in progress_bar(exp_v.Test_Sets.items()):
            sleep(1)

    def named() -> None:
        """
        :rtype: None
        """
        from time import sleep

        list_items = [2.13, 8921.9123, 923]

        for a in progress_bar(list_items):
            sleep(1)

    def named_time() -> None:
        """
        :rtype: None
        """
        from time import sleep

        list_items = [2.13, 8921.9123, 923]

        for a in progress_bar(list_items, timestamp_mode=TimestampModeEnum.prefix):
            sleep(1)

    def send_example() -> None:
        """
        :rtype: None
        """
        from itertools import count

        pb = progress_bar(count(), timestamp_mode=TimestampModeEnum.none)
        next(pb)
        for a in range(100000):
            pb.send(f"step_{a}")

    def send_example_time() -> None:
        """
        :rtype: None
        """
        from itertools import count

        pb = progress_bar(count(), timestamp_mode=TimestampModeEnum.postfix)
        next(pb)
        for a in range(100000):
            pb.send(f"step_{a}")

    # dsad3123()
    # asd21sa()
    # dict_items()
    # send_example()
    # named()
    named_time()
    # send_example_time()



from enum import Enum
from typing import Iterable

from sorcery import assigned_names

__all__ = ["ExtensionEnum", "match_return_code"]

ESC_CHAR = chr(27)


def match_return_code(ret_val, chars: Iterable[str] = ("q", ESC_CHAR)):
    return any(ret_val & 0xFF == ord(c) for c in chars)


class ExtensionEnum(Enum):
    png, exr = assigned_names()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/06/2020
           """

import torch

from warg import AlsoDecorator

__all__ = ["DefaultTypeSession"]


# torch.set_default_tensor_type('torch.cuda.FloatTensor') # Legacy


class DefaultTypeSession(AlsoDecorator):
    """
    # speed up evaluating after training finished"""

    def __init__(self, dtype: torch.dtype = torch.float32, no_side_effect: bool = True):
        self._dtype = dtype
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_state = torch.get_default_dtype()

    def __enter__(self):
        torch.set_default_dtype(self._dtype)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            torch.set_default_dtype(self.prev_state)
        else:
            torch.set_default_dtype(torch.float32)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-03-2021
           """

__all__ = [
    "round_tight_bbox",
    "opaque_round_tight_bbox",
    "semi_opaque_round_tight_bbox",
    "lt_ann_transform",
    "lb_ann_transform",
    "rt_ann_transform",
    "rb_ann_transform",
    "arc_arrow",
]

round_tight_bbox = dict(
    boxstyle="round,pad=.2,rounding_size=.2", fc="1.0", alpha=0.5, ec="0.9"
)
opaque_round_tight_bbox = round_tight_bbox.copy()
opaque_round_tight_bbox["alpha"] = 1.0
semi_opaque_round_tight_bbox = round_tight_bbox.copy()
semi_opaque_round_tight_bbox["alpha"] = 0.75
lt_ann_transform = dict(xytext=(1.4, -0.8), ha="left", va="top", rotation=-35)
lb_ann_transform = dict(xytext=(1.4, 0.8), ha="left", va="bottom", rotation=35)
rt_ann_transform = dict(xytext=(-1.4, -0.8), ha="right", va="top", rotation=35)
rb_ann_transform = dict(xytext=(-1.4, 0.8), ha="right", va="bottom", rotation=-35)
arc_arrow = (dict(arrowstyle="->", connectionstyle="arc3"),)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

from .platform_selection import *
from .cuda_device import *
from .screen_resolution import *
from .resource_utilities import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["EmbedWriterMixin"]

from typing import Any, Sequence


class EmbedWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' embeds/projections(2d,3d) for interactive visualisation"""

    @abstractmethod
    def embed(
        self,
        tag: str,
        response: Sequence,
        metadata: Any = None,
        label_img: Any = None,  # raster grid / image / numpy.array
        step: int = None,
        **kwargs
    ) -> None:
        """
            eg. visualising for projections in lower dimensional space

            :param response:
            :param metadata:
            :param label_img:
        :param tag:
        :type tag:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "{user}"
__doc__ = r"""

           Created on {date}
           """

from enum import Enum

from sorcery import assigned_names

__all__ = ["ReductionMethodEnum"]


class ReductionMethodEnum(Enum):
    """ """

    none, mean, sum = assigned_names()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 29/07/2020
           """

from .checking import *
from .flow import *
from .grad_trace import *
from .guided import *



from enum import Enum
from typing import Any

import cv2
from sorcery import assigned_names

from warg import ceil_odd

__all__ = ["noise_filter", "NoiseFilterMethodEnum"]


class NoiseFilterMethodEnum(Enum):
    none, median_blur, bilateral_filter, gaussian_blur = assigned_names()


def noise_filter(
    img: Any,
    method: NoiseFilterMethodEnum = NoiseFilterMethodEnum.bilateral_filter,
    **kwargs
) -> Any:
    method = NoiseFilterMethodEnum(method)

    if method == NoiseFilterMethodEnum.none:
        return img

    ksize = kwargs.get("ksize", max(ceil_odd(max(*(img.shape)) // 100), 5))
    if method == NoiseFilterMethodEnum.median_blur:
        return cv2.medianBlur(img, ksize=ksize)
    elif method == NoiseFilterMethodEnum.bilateral_filter:
        return cv2.bilateralFilter(
            img,
            d=kwargs.get("d", 11),
            sigmaColor=kwargs.get("sigmaColor", 17),
            sigmaSpace=kwargs.get("sigmaSpace", 17),
        )
    elif method == NoiseFilterMethodEnum.gaussian_blur:

        return cv2.GaussianBlur(
            img,
            ksize=(ksize, ksize),
            sigmaX=kwargs.get("sigmaX", 5),
            borderType=kwargs.get("borderType", None),
        )

    raise NoiseFilterMethodEnum



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-05-2021
           """

import random
from functools import partial

import numpy
import pandas

from draugr import indent_lines
from draugr.pandas_utilities.formatting import pandas_mean_std_bold_formatter
from warg import Number, drop_unused_kws, passes_kws_to

__all__ = [
    "pandas_mean_std",
    "pandas_mean_std_to_str",
    "pandas_format_bold_max_row_latex",
    "pandas_to_latex_clean",
]


def pandas_mean_std(df: pandas.DataFrame, group_by: str) -> pandas.DataFrame:
    """

    :param df:
    :param group_by:
    :return:
    """
    return df.groupby(group_by).agg([numpy.mean, numpy.std])


def pandas_mean_std_to_str(
    df: pandas.DataFrame,
    *,
    mean_col: str = "mean",
    std_col: str = "std",
    precision: int = 3,
    level=1,
    axis=1,
) -> pandas.DataFrame:
    """
    latex \pm plus minus
    :param level:
    :param axis:
    :param std_col:
    :param mean_col:
    :param df:
    :param precision:
    :return:
    """
    return (
        df.xs(mean_col, axis=axis, level=level).round(precision).astype(str)
        + " \pm "
        + df.xs(std_col, axis=axis, level=level).round(precision).astype(str)
    )


def pandas_mean_std_latex_tabular(
    df: pandas.DataFrame,
    group_by: str,
    *,
    precision: int = 3,
    header_rotation: Number = 0,
) -> str:
    """

    :param df:
    :param group_by:
    :param precision:
    :param header_rotation:
    :return:
    """
    mean_std = pandas_mean_std(df, group_by)
    mean_std_str = pandas_mean_std_to_str(mean_std, precision=precision)

    return pandas_format_bold_max_column_latex(
        mean_std, mean_std_str, precision=precision, header_rotation=header_rotation
    )


def pandas_format_bold_max_column_latex(
    max_provider_df: pandas.DataFrame,
    entry_provider_df: pandas.DataFrame,
    *,
    precision: int = 3,
    header_rotation: Number = 0,
    max_colwidth: int = 1000,
):
    """

    :param max_provider_df:
    :param entry_provider_df:
    :param precision:
    :param header_rotation:
    :param max_colwidth:
    :return:
    """
    formatters = {
        column[0]: partial(
            pandas_mean_std_bold_formatter,
            value=max_provider_df[column].max(),
            precision=precision,
        )
        for column in max_provider_df.columns
        if "mean" in column
    }
    return pandas_to_latex_clean(
        entry_provider_df,
        header_rotation=header_rotation,
        max_colwidth=max_colwidth,
        formatters=formatters,
        precision=precision,
    )


def pandas_format_bold_max_row_latex(
    max_provider_df: pandas.DataFrame,
    entry_provider_df: pandas.DataFrame,
    *,
    precision: int = 3,
    header_rotation: Number = 0,
    max_colwidth=1000,
):
    """

    :param max_provider_df:
    :param entry_provider_df:
    :param precision:
    :param header_rotation:
    :param max_colwidth:
    :return:
    """
    formatters = {
        column[0]: partial(
            pandas_mean_std_bold_formatter,
            value=max_provider_df[column].max(),
            precision=precision,
        )
        for column in max_provider_df.columns
        if "mean" in column
    }

    return pandas_to_latex_clean(
        entry_provider_df,
        header_rotation=header_rotation,
        max_colwidth=max_colwidth,
        formatters=formatters,
    )


def pandas_to_latex_clean(
    entry_provider_df: pandas.DataFrame,
    *,
    header_rotation: Number = 0,
    precision=3,
    max_colwidth=1000,
    formatters=None,
) -> str:
    """

    :param entry_provider_df:
    :param header_rotation:
    :param precision:
    :param max_colwidth:
    :param formatters:
    :return:
    """
    if True:
        if not isinstance(entry_provider_df.columns, pandas.MultiIndex):
            entry_provider_df.columns = entry_provider_df.columns.str.replace("_", "\_")
        else:
            entry_provider_df.columns = [
                _name.replace("_", "\_") if _name else _name
                for _name in entry_provider_df.columns
            ]
            """
entry_provider_df.columns = pandas.MultiIndex(levels=[
[col.replace('_', '\_') for col in lvl]
for lvl in entry_provider_df.columns.levels
],
codes=entry_provider_df.columns.codes,
names=entry_provider_df.columns.names
)
"""

        if not isinstance(entry_provider_df.index, pandas.MultiIndex):
            entry_provider_df.index = entry_provider_df.index.str.replace("_", "\_")
        else:
            entry_provider_df.index = pandas.MultiIndex(
                levels=[
                    [col.replace("_", "\_") for col in lvl]
                    for lvl in entry_provider_df.index.levels
                ],
                codes=entry_provider_df.index.codes,
                names=entry_provider_df.index.names,
            )

        entry_provider_df.columns.names = [
            _name.replace("_", "\_") if _name else _name
            for _name in entry_provider_df.columns.names
        ]
        entry_provider_df.index.names = [
            _name.replace("_", "\_") if _name else _name
            for _name in entry_provider_df.index.names
        ]

    if formatters is None:
        formatters = {
            column: partial(
                pandas_mean_std_bold_formatter, value=None, precision=precision
            )
            for column in entry_provider_df.columns
        }

    if header_rotation:
        header = [
            f"\\rotatebox{{{header_rotation}}}" + "{" + "\_".join(c.split("_")) + "}"
            for c in entry_provider_df.columns
        ]
    else:
        header = True

    with pandas.option_context("max_colwidth", max_colwidth):
        return entry_provider_df.to_latex(
            index=True,
            index_names=[*entry_provider_df.index.names],
            escape=False,
            formatters=formatters,
            header=header,
        )  # .replace('textbackslash ', '').replace('\$', '$')


@drop_unused_kws
@passes_kws_to(pandas_mean_std_latex_tabular)
def pandas_mean_std_latex_table(
    df: pandas.DataFrame,
    group_by: str,
    tab_label: str,
    *,
    truncate_n_tabular_symbols: int = 2,
    **kwargs,
) -> str:
    """

    :param df:
    :param group_by:
    :param tab_label:
    :param truncate_n_tabular_symbols:
    :param kwargs:
    :return:
    """
    return (
        f"""\\begin{{table}}
  \caption{{{tab_label.replace('_', ' ')}}}
  \label{{tab:{tab_label}}}
  \centering
"""
        + indent_lines(pandas_mean_std_latex_tabular(df, group_by, **kwargs))[
            :-truncate_n_tabular_symbols
        ]
        + """\end{table}"""
    )


'''
def pandas_mean_std_latex_table8(df: pandas.DataFrame,
                                group_by: str,
                                precision: int = 3,
                                header_rotation: Number = 0) -> str:
  out = df.groupby(group_by).agg([numpy.mean, numpy.std])

  def bold_formatter(x, value):
    """Format a number in bold when (almost) identical to a given value.

    Args:
        x: Input number.

        value: Value to compare x with.

        num_decimals: Number of decimals to use for output format.

    Returns:
        String converted output.

    """
    # Consider values equal, when rounded results are equal
    # otherwise, it may look surprising in the table where they seem identical
    a = float(x.split('\pm')[0])

    if round(a, precision) == round(value, precision):
      return f"$\\mathbf{{{x}}}$"
    else:
      return f"${x}$"

  asds = {column[0]:partial(bold_formatter,
                            value=out[column].max()) for column in out.columns if 'mean' in column}

  out = (out.xs('mean', axis=1, level=1).round(precision).astype(str) + ' \pm ' + out.xs('std', axis=1, level=1).round(precision).astype(str))

  return out.to_latex(index=True,
                      escape=False,
                      formatters=asds,
                      header=[f'\\rotatebox{{{header_rotation}}}' + '{' + "\_".join(c.split("_")) + '}' for c in out.columns]
                      )  # .replace('textbackslash ', '').replace('\$', '$')
'''

if __name__ == "__main__":

    def asuhda() -> None:
        """
        :rtype: None
        """
        isjda = "dx_sijdai_iahjdaw-_sdioja_sakodwada_soakd_aoskdiojwd_s"
        df = pandas.DataFrame(
            numpy.random.randint(0, 100, size=(15, 3)),
            columns=("roc_asdoj", "au_c", "la"),
        )
        df["c_asd"] = [
            isjda[random.randint(0, 15) : random.randint(5, 30)] for _ in range(len(df))
        ]

        def asodjiasj():
            """ """
            print(pandas_mean_std(df, "c_asd"))

        def asidj():
            """ """
            print(pandas_mean_std_latex_table(df, "c_asd", "some_table"))

        asodjiasj()
        asidj()

    asuhda()



from enum import Enum


class TaskTriggerEnum(Enum):
    """ """

    TASK_TRIGGER_EVENT = 0
    TASK_TRIGGER_TIME = 1
    TASK_TRIGGER_DAILY = 2
    TASK_TRIGGER_WEEKLY = 3
    TASK_TRIGGER_MONTHLY = 4
    TASK_TRIGGER_MONTHLYDOW = 5
    TASK_TRIGGER_IDLE = 6
    TASK_TRIGGER_REGISTRATION = 7
    TASK_TRIGGER_BOOT = 8
    TASK_TRIGGER_LOGON = 9
    TASK_TRIGGER_SESSION_STATE_CHANGE = 11



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15-02-2021
           """

from .annotation import *
from .cyclers import *
from .hatching import *
from .lines import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

import collections
import hashlib
import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, OrderedDict, Sequence

import numpy

__all__ = ["SplitEnum", "SplitIndexer", "train_valid_test_split", "select_split"]

from sorcery import assigned_names


class SplitEnum(Enum):
    """
    Split Enum class for selecting splits"""

    (training, validation, testing) = assigned_names()


class SplitIndexer:
    """Splits dataset in to 3 parts based on percentages, returns indices for the data set sequence"""

    default_split_names = {i: i.value for i in SplitEnum}

    def __init__(
        self,
        dataset_length: int,
        training: float = 0.7,
        validation: float = 0.2,
        testing: float = 0.1,
    ):
        self.total_num = dataset_length
        splits = numpy.array([training, validation, testing])
        self.normalised_split = splits / sum(splits)
        (
            self.training_percentage,
            self.validation_percentage,
            self.testing_percentage,
        ) = self.normalised_split
        self.training_num, self.validation_num, self.testing_num = self.unnormalised(
            dataset_length
        )

    def shuffled_indices(self) -> dict:
        """ """
        split_indices = numpy.random.permutation(self.total_num).tolist()

        return {
            SplitEnum.training: self.select_train_indices(split_indices),
            SplitEnum.validation: self.select_validation_indices(split_indices),
            SplitEnum.testing: self.select_testing_indices(split_indices),
        }

    def select_train_indices(self, ind: Sequence) -> Sequence:
        """ """
        return ind[: self.training_num]

    def select_validation_indices(self, ind: Sequence) -> Sequence:
        """ """
        if self.validation_num:
            if self.testing_num:
                return ind[self.training_num : -self.testing_num]
            return ind[self.training_num :]
        return []

    def select_testing_indices(self, ind: Sequence) -> Sequence:
        """ """
        if self.testing_num:
            return ind[-self.testing_num :]
        return []

    def unnormalised(self, num: int, floored: bool = True) -> numpy.ndarray:
        """

        :param num:
        :type num:
        :param floored:
        :type floored:
        :return:
        :rtype:"""
        unnorm = self.normalised_split * num
        if floored:
            unnorm = numpy.floor(unnorm)
        return unnorm.astype(int)

    def __repr__(self) -> str:
        return str(
            {k: n for k, n in zip(self.default_split_names, self.normalised_split)}
        )

    def select_shuffled_split_indices(self, split: SplitEnum, seed: int = 0) -> object:
        """ """
        numpy.random.seed(seed)
        split_indices = numpy.random.permutation(self.total_num).tolist()

        if split == SplitEnum.training:
            return self.select_train_indices(split_indices)
        elif split == SplitEnum.validation:
            return self.select_validation_indices(split_indices)
        elif split == SplitEnum.testing:
            return self.select_testing_indices(split_indices)
        elif split is None:
            return split_indices
        raise NotImplementedError


def train_valid_test_split(
    categories: Dict[str, Iterable[Path]],
    *,
    validation_percentage: float = 15,  # TODO: ACCEPT AND SQUEEZE ZERO-HUNDRED TO ZERO-ONE range!
    testing_percentage: float = 0,
    verbose: bool = False,
) -> OrderedDict:
    """
    Magic hashing

    :param verbose:
    :type verbose:
    :param categories:
    :param testing_percentage:
    :param validation_percentage:
    :return:"""
    result = collections.OrderedDict()

    if verbose:
        print(categories)

    for c, vs in categories.items():
        training_images = []
        testing_images = []
        validation_images = []

        for file_name in vs:
            b_rep = bytes(re.sub(r"_nohash_.*$", "", f"{c}{file_name.name}"), "utf8")
            percentage_hash = (
                int(hashlib.sha1(b_rep).hexdigest(), 16) % (sys.maxsize + 1)
            ) * (100.0 / sys.maxsize)
            if percentage_hash < validation_percentage + testing_percentage:
                if percentage_hash < testing_percentage:
                    testing_images.append(file_name)
                else:
                    validation_images.append(file_name)
            else:
                training_images.append(file_name)

        result[c] = {
            SplitEnum.training: training_images,
            SplitEnum.validation: validation_images,
            SplitEnum.testing: testing_images,
        }

    return result


def select_split(
    data_cat_split: Dict[Any, Dict[SplitEnum, Sequence]],
    split: SplitEnum,
    verbose: bool = False,
) -> Dict[Any, Sequence]:
    """

    :param verbose:
    :type verbose:
    :param data_cat_split:
    :type data_cat_split:
    :param split:
    :type split:
    :return:
    :rtype:"""
    data = {k: [] for k in data_cat_split.keys()}
    if verbose:
        print(data_cat_split)
    for k, v in data_cat_split.items():
        if verbose:
            print(v[split])
        for item in v[split]:
            data[k].append(item)
    return data


if __name__ == "__main__":

    def asd():
        split_by_p = SplitIndexer(100)
        print(split_by_p.default_split_names)
        print(split_by_p.shuffled_indices())
        print(split_by_p.select_shuffled_split_indices(SplitEnum.training))
        a = split_by_p.select_shuffled_split_indices(None)
        print(a, len(a))

    def uihsad():
        a = None
        if a:
            a = SplitEnum(a)
        print(a)

    uihsad()



import os
import subprocess
from pathlib import Path

from apppath import ensure_existence

__all__ = ["spilt_video"]


def spilt_video(
    video_path: Path,
    start_time="01:40:00",
    stop_time="01:50:00",
    split_dir: Path = None,
    ffmpeg_path: Path = "ffmpeg",
) -> None:
    if split_dir is None:
        split_dir = ensure_existence(video_path.parent / "split")

    if not split_dir.is_dir():
        ensure_existence(split_dir)

    ext = video_path.suffix
    name = video_path.stem

    while os.path.exists(str(split_dir / (name + ext))):
        name = name + "-1"

    subprocess.call(
        [
            ffmpeg_path,
            "-ss",
            start_time,
            "-to",
            stop_time,
            "-accurate_seek",
            "-i",
            video_path,
            "-vcodec",
            "copy",
            "-acodec",
            "copy",
            "-avoid_negative_ts",
            "1",
            str(split_dir / (name + ext)),
            "-y",
        ]
    )


if __name__ == "__main__":
    spilt_video(
        Path.home() / "Downloads" / "brandt.mp4",
        ffmpeg_path=Path.home()
        / "Downloads"
        / "ffmpeg-5.0-essentials_build"
        / "ffmpeg-5.0-essentials_build"
        / "bin"
        / "ffmpeg.exe",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/03/2020
           """
__all__ = ["get_host_ip", "find_unclaimed_port", "is_port_in_use"]


def get_host_ip() -> str:
    """Get host ip.

    Returns:
    str: The obtained ip. UNKNOWN if failed."""
    import socket

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
    except:
        ip = "UNKNOWN"
    finally:
        s.close()

    return ip


def find_unclaimed_port() -> int:
    """
    # NOTE: there is still a chance the port could be taken by other processes.
    :return:
    :rtype:"""
    import socket

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(
        ("", 0)
    )  # Binding to port 0 will cause the OS to find an available port for us
    port = sock.getsockname()[1]
    sock.close()
    return port


def is_port_in_use(port: int) -> bool:
    """

    :param port_:
    :return:
    """
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


if __name__ == "__main__":
    HOST_IP = get_host_ip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/03/2020
           """

from functools import partial
from typing import Callable, Iterable, Optional

import cv2

__all__ = ["frame_generator", "to_rgb"]

from warg import identity

to_rgb = partial(cv2.cvtColor, code=cv2.COLOR_BGR2RGB)


def frame_generator(
    video_stream: cv2.VideoCapture,
    coder: Optional[Callable] = identity,
) -> Iterable:
    """

    :param video_stream:
    :param coder:"""
    if coder is None:
        coder = identity
    while video_stream.isOpened():
        success, frame = video_stream.read()
        if success:
            yield coder(frame)
        else:
            break


if __name__ == "__main__":

    def asd() -> None:
        """
        :rtype: None
        """
        from draugr.opencv_utilities.windows.image import show_image
        from draugr.tqdm_utilities import progress_bar

        for image in progress_bar(frame_generator(cv2.VideoCapture(0))):
            if show_image(image, "frame", wait=1):
                break  # esc to quit

    asd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Iterable, Sequence, Union

import numpy
import torch
import torchvision
from PIL.Image import Image

__author__ = "Christian Heider Nielsen"
__doc__ = ""
__all__ = ["to_tensor"]

from draugr.torch_utilities.tensors.types import numpy_to_torch_dtype


# @passes_kws_to(torch.Tensor.to)
def to_tensor(
    obj: Union[torch.Tensor, numpy.ndarray, Iterable, Sequence, int, float],
    dtype: Union[
        torch.dtype, object
    ] = None,  # if None, torch.float or equivalent numpy.dtype is used, can be used to force dtype
    device: Union[str, torch.device] = "cpu",
    **kwargs
) -> torch.Tensor:
    """

    :param obj:
    :param dtype:
    :param device:
    :param kwargs:
    :return:"""

    if dtype is None:
        use_dtype = torch.float
    else:
        if dtype == float:
            dtype = torch.float  # TODO: LOOKUP TABLE OF TYPES
        use_dtype = dtype

    # torch.as_tensor()
    if torch.is_tensor(obj):
        if dtype is None:
            use_dtype = obj.dtype
        return obj.to(dtype=use_dtype, device=device, **kwargs)

    if isinstance(obj, Image):
        return torchvision.transforms.functional.to_tensor(obj)

    if isinstance(obj, numpy.ndarray):
        if dtype is None:
            use_dtype = numpy_to_torch_dtype(obj.dtype)
        if torch.is_tensor(obj[0]) and len(obj[0].size()) > 0:
            return torch.stack(obj.tolist()).to(
                dtype=use_dtype, device=device, **kwargs
            )
        return torch.from_numpy(obj).to(dtype=use_dtype, device=device, **kwargs)

    if not isinstance(obj, Sequence):
        if isinstance(obj, set):
            obj = [*obj]
        else:
            obj = [obj]
    elif not isinstance(obj, list) and isinstance(obj, Iterable):
        obj = [*obj]

    if isinstance(obj, list):
        if torch.is_tensor(obj[0]) and len(obj[0].size()) > 0:
            return torch.stack(obj).to(dtype=use_dtype, device=device, **kwargs)
        elif isinstance(obj[0], list):
            obj = [to_tensor(o) for o in obj]
            return torch.stack(obj).to(dtype=use_dtype, device=device, **kwargs)
        elif isinstance(
            obj[0], numpy.ndarray
        ):  # TODO: ASSERT ALL SAME DTYPE and shape and is array at all
            obj = numpy.stack(obj)
            # obj = numpy.ndarray(obj) # does not stack!
            # obj = [to_tensor(o) for o in obj]
            # return torch.stack(obj).to(dtype=use_dtype, device=device, **kwargs)

    return torch.tensor(obj, dtype=use_dtype, device=device, **kwargs)


if __name__ == "__main__":
    print(to_tensor(1))
    print(to_tensor(2.0))
    print(to_tensor([0.5, 0.5]))
    print(to_tensor([[0.5, 0.5]]))
    print(to_tensor((0.5, 0.5)))
    print(to_tensor(range(10)))
    print(to_tensor(torch.from_numpy(numpy.array([0.5, 0.5]))))
    a = torch.arange(0, 10)
    print(to_tensor(a))
    print(to_tensor([a, a]))
    print(to_tensor((a, a)))

    print(to_tensor((torch.zeros((2, 2)), (torch.zeros((2, 2))))).shape)
    print(to_tensor(([torch.zeros((2, 2)), torch.zeros((2, 2))])).shape)
    print(to_tensor(([torch.zeros((2, 2))], [torch.zeros((2, 2))])).shape)
    print(to_tensor(([[[torch.zeros((2, 2))]]], [[[torch.zeros((2, 2))]]])).shape)
    print(to_tensor(([[torch.zeros((2, 2))], [torch.zeros((2, 2))]])).shape)
    print(to_tensor([(torch.zeros((2, 2))), (torch.zeros((2, 2)))]).shape)
    print(to_tensor({(torch.zeros((2, 2))), (torch.zeros((2, 2)))}).shape)
    print(to_tensor(((torch.zeros((2, 2))), (torch.zeros((2, 2))))).shape)
    print(to_tensor([[[[torch.zeros((2, 2))]], [[torch.zeros((2, 2))]]]]).shape)

    print(
        to_tensor(
            (
                numpy.zeros((2, 2)),
                numpy.zeros((2, 2)),
                numpy.zeros((2, 2)),
                numpy.zeros((2, 2)),
            )
        )
    )

    print(
        (
            *[
                to_tensor(a, device="cpu")
                for a in [
                    numpy.zeros((2, 2)),
                    numpy.zeros((2, 2)),
                    numpy.zeros((2, 2)),
                    numpy.zeros((2, 2)),
                ]
            ],
        )
    )

    print(
        to_tensor(
            [
                to_tensor(numpy.zeros((2, 2))),
                to_tensor(numpy.zeros((2, 2))),
                to_tensor(numpy.zeros((2, 2))),
                to_tensor(numpy.zeros((2, 2))),
            ]
        )
    )

    print(
        (
            *[
                to_tensor(a, device="cpu")
                for a in [
                    to_tensor(numpy.zeros((2, 2))),
                    to_tensor(numpy.zeros((2, 2))),
                    to_tensor(numpy.zeros((2, 2))),
                    to_tensor(numpy.zeros((2, 2))),
                ]
            ],
        )
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 22-01-2021
           """

from typing import Sequence

from matplotlib import pyplot

from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer
from warg import passes_kws_to

__all__ = ["ImageStreamPlot"]


class ImageStreamPlot(MplDrawer):
    """ """

    @passes_kws_to(MplDrawer.__init__)
    @passes_kws_to(pyplot.imshow)
    def __init__(
        self, image, title: str = "", render: bool = True, **kwargs
    ):  # size:Tuple[int,...],
        super().__init__(render=render, **kwargs)

        if not render:
            return

        self.fig = pyplot.figure(figsize=(4, 4))
        self.im = pyplot.imshow(image, **kwargs)

        pyplot.title(title)
        pyplot.tight_layout()
        cid = self.fig.canvas.mpl_connect("key_press_event", self.close)

    def _draw(self, data: Sequence):
        """

        :param data:
        :return:"""
        self.im.set_data(data)

    @staticmethod
    def close(event):
        """ """
        if event.key == "q":
            pyplot.close(event.canvas.figure)
            raise StopIteration


if __name__ == "__main__":

    def asdasf() -> None:
        """
        :rtype: None
        """
        import cv2
        from draugr.opencv_utilities import frame_generator
        from draugr.tqdm_utilities import progress_bar
        from functools import partial
        from draugr.opencv_utilities import AsyncVideoStream

        with AsyncVideoStream() as vc:
            coder = partial(cv2.cvtColor, code=cv2.COLOR_BGR2RGB)
            with ImageStreamPlot(coder(next(vc))) as s:
                for i in progress_bar(
                    frame_generator(vc, coder=coder), auto_total_generator=False
                ):
                    s.draw(i)

    asdasf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from enum import Enum
from typing import Union

import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/11/2019
           """

from sorcery import assigned_names

GLOBAL_DEVICE: torch.device = None

__all__ = [
    "global_torch_device",
    "select_cuda_device",
    "get_gpu_usage_mb",
    "auto_select_available_cuda_device",
    "set_global_torch_device",
    "torch_clean_up",
    "TorchDeviceEnum",
]


class TorchDeviceEnum(Enum):
    (cpu, cuda, vulkan) = assigned_names()


def global_torch_device(
    device_preference: Union[bool, str, TorchDeviceEnum] = None,
    override: torch.device = None,
    verbose: bool = False,
) -> torch.device:
    """

    first time call stores to device for global reference, later call must explicitly manually override!

    :param verbose:
    :type verbose:
    :param device_preference:
    :type device_preference:
    :param override:
    :type override:
    :return:
    :rtype:"""
    global GLOBAL_DEVICE

    if override is not None:
        GLOBAL_DEVICE = override
        if verbose:
            print(f"Overriding global torch device to {override}")
    elif device_preference is not None:
        if isinstance(device_preference, bool):
            if torch.is_vulkan_available() and device_preference:
                d_ = TorchDeviceEnum.vulkan
            elif torch.cuda.is_available() and device_preference:
                d_ = TorchDeviceEnum.cuda
            else:
                d_ = TorchDeviceEnum.cpu

            d = torch.device(d_.value)
        elif isinstance(device_preference, TorchDeviceEnum):
            d = torch.device(device_preference.value)
        elif isinstance(device_preference, str):
            d = torch.device(device_preference)
        else:
            raise TypeError("not bool or str")
        if GLOBAL_DEVICE is None:
            GLOBAL_DEVICE = d
        return d
    elif GLOBAL_DEVICE is None:
        if torch.is_vulkan_available():
            d_ = TorchDeviceEnum.vulkan
        elif torch.cuda.is_available():
            d_ = TorchDeviceEnum.cuda
        else:
            d_ = TorchDeviceEnum.cpu
        GLOBAL_DEVICE = torch.device(d_.value)

    return GLOBAL_DEVICE


def set_global_torch_device(device: torch.device) -> None:
    """

    :param device:
    :return:"""
    global GLOBAL_DEVICE
    GLOBAL_DEVICE = device


def select_cuda_device(cuda_device_idx: int) -> torch.device:
    """

    :param cuda_device_idx:
    :type cuda_device_idx:
    :return:
    :rtype:"""
    num_cuda_device = torch.cuda.device_count()
    assert num_cuda_device > 0
    assert cuda_device_idx < num_cuda_device
    if 0 <= cuda_device_idx < num_cuda_device:
        return torch.device(f"cuda:{cuda_device_idx}")


def get_gpu_usage_mb():
    """

    :return:
    :rtype:"""

    import subprocess

    """Get the current gpu usage.

Returns
-------
usage: dict
Keys are device ids as integers.
Values are memory usage as integers in MB.
"""
    result = subprocess.check_output(
        ["nvidia-smi", "--query-gpu=memory.used", "--format=csv,nounits,noheader"]
    ).decode("utf-8")
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split("\n")]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map


def torch_clean_up() -> None:
    r"""**Destroy cuda state by emptying cache and collecting IPC.**

    Consecutively calls `torch.cuda.empty_cache()` and `torch.cuda.ipc_collect()`."""

    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()


def auto_select_available_cuda_device(
    expected_memory_usage_mb: int = 1024,
) -> torch.device:
    r"""
    Auto selects the device with highest compute capability and with the requested memory available

    :param expected_memory_usage_mb:
    :type expected_memory_usage_mb:
    :return:
    :rtype:"""

    num_cuda_device = torch.cuda.device_count()
    # TODO: torch.vulkan.device_count() variant in the future
    assert num_cuda_device > 0
    """
print(torch.cuda.cudart())
print(torch.cuda.memory_snapshot())
torch.cuda.memory_cached(dev_idx),
torch.cuda.memory_allocated(dev_idx),
torch.cuda.max_memory_allocated(dev_idx),
torch.cuda.max_memory_cached(dev_idx),
torch.cuda.get_device_name(dev_idx),
torch.cuda.get_device_properties(dev_idx),
torch.cuda.memory_stats(dev_idx)
"""
    preferred_idx = None
    highest_capability = 0
    for dev_idx, usage in enumerate(get_gpu_usage_mb().values()):
        cuda_capability = float(
            ".".join([str(x) for x in torch.cuda.get_device_capability(dev_idx)])
        )
        if expected_memory_usage_mb:
            total_mem = (
                torch.cuda.get_device_properties(dev_idx).total_memory // 1000 // 1000
            )
            if expected_memory_usage_mb < total_mem - usage:
                if cuda_capability > highest_capability:
                    highest_capability = cuda_capability
                    preferred_idx = dev_idx
        else:
            if cuda_capability > highest_capability:
                highest_capability = cuda_capability
                preferred_idx = dev_idx
    if preferred_idx is None:
        raise FileNotFoundError(
            f"No device with {expected_memory_usage_mb} mb memory found"
        )
    return select_cuda_device(preferred_idx)


if __name__ == "__main__":

    def stest_override() -> None:
        """
        :rtype: None
        """
        print(global_torch_device(verbose=True))
        print(
            global_torch_device(
                override=global_torch_device(device_preference=False, verbose=True),
                verbose=True,
            )
        )
        print(global_torch_device(verbose=True))
        print(global_torch_device(device_preference=True))
        print(global_torch_device())
        print(
            global_torch_device(
                override=global_torch_device(device_preference=True, verbose=True)
            )
        )
        print(global_torch_device())

    def a() -> None:
        """
        :rtype: None
        """
        print(global_torch_device())
        print(auto_select_available_cuda_device())

    def b() -> None:
        """
        :rtype: None
        """
        print(global_torch_device(TorchDeviceEnum.vulkan))

    # stest_override()

    print(global_torch_device(False).type)
    b()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

from pathlib import Path

from matplotlib import pyplot
from matplotlib.image import imread

__all__ = ["display_depth_map"]


def display_depth_map(
    data_set_directory: Path = Path.home()
    / "Datasets"
    / "neodroid"
    / "depth"
    / "80.png",
):
    """

    :param data_set_directory:
    :type data_set_directory:"""
    # img = Image.open(data_set_directory + file_name).convert('LA')
    # img_array =numpy.asarray(img)
    # print(img_array.shape)

    img = imread(data_set_directory)
    img = img[..., 0]
    img = img / 255

    pyplot.imshow(img, cmap=pyplot.get_cmap("gray"))
    pyplot.show()


if __name__ == "__main__":
    display_depth_map()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021. Lorem ipsum dolor sit amet, consectetur adipiscing elit.
#  Morbi non lorem porttitor neque feugiat blandit. Ut vitae ipsum eget quam lacinia accumsan.
#  Etiam sed turpis ac ipsum condimentum fringilla. Maecenas magna.
#  Proin dapibus sapien vel ante. Aliquam erat volutpat. Pellentesque sagittis ligula eget metus.
#  Vestibulum commodo. Ut rhoncus gravida arcu.

import pickle
import time
from abc import ABC, abstractmethod
from typing import Any, Iterable

from cloudpickle import cloudpickle

__author__ = "Christian Heider Nielsen"

import multiprocessing
import queue

__all__ = ["CloudPickleBase", "PooledQueueTask", "PooledQueueProcessor"]


class CloudPickleBase(object):
    """
    Uses cloudpickle to serialize contents (otherwise multiprocessing tries to use pickle)
    :param x: (Any) the variable you wish to wrap for pickling with cloudpickle"""

    def __init__(self, x: Any):
        self._x = x

    def __getstate__(self):
        return cloudpickle.dumps(self._x)

    def __setstate__(self, x):
        self._x = pickle.loads(x)

    def __call__(self, *args, **kwargs):
        return self._x(*args, **kwargs)


class PooledQueueTask(ABC):
    """
    Pooled queue task interface"""

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

    @abstractmethod
    def call(self, *args, **kwargs) -> Any:
        """

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:"""
        raise NotImplemented


class PooledQueueProcessor(object):
    """
    This is a workaround of Pythons extremely slow interprocess communication pipes.
    The ideal solution would be to use a multiprocessing.queue, but it apparently communication is band
    limited.
    This solution has processes complete tasks (batches) and a thread add the results to a queue.queue."""

    def __init__(
        self,
        func,
        args: Iterable = (),
        kwargs=None,
        max_queue_size=100,
        n_proc=None,
        max_tasks_per_child=None,
        fill_at_construction=True,
        blocking=True,
    ):

        if kwargs is None:
            kwargs = {}

        self._max_queue_size = max_queue_size
        if isinstance(func, type):
            func = func()
        self._func = CloudPickleBase(func)
        self.args = args
        self.kwargs = kwargs
        self.blocking = blocking
        if max_tasks_per_child is None:
            max_tasks_per_child = max_queue_size // 4

        if n_proc is None:
            n_proc = multiprocessing.cpu_count()

        self._queue = queue.Queue(maxsize=max_queue_size)
        self._pool = multiprocessing.Pool(n_proc, maxtasksperchild=max_tasks_per_child)

        if fill_at_construction:
            self.fill()

    def fill(self) -> None:
        """
        fill queue"""
        for i in range(self._max_queue_size):
            self.maybe_fill()

    def close(self) -> None:
        """
        close pool"""
        self._pool.close()
        self._pool.join()

    def terminate(self) -> None:
        """
        terminate pool"""
        self._pool.terminate()
        self._pool.join()

    def maybe_fill(self) -> None:
        """
        fill queue if not full"""
        if self.queue_size < self._max_queue_size:  # and not self._queue.full():
            self._pool.apply_async(
                self._func, self.args, self.kwargs, self.put, self.raise_error
            )

    @property
    def queue_size(self) -> int:
        """

        :return:
        :rtype:"""
        return self._queue.qsize()

    def put(self, res) -> None:
        """

        :param res:
        :type res:"""
        self._queue.put(res)

    def raise_error(self, excptn) -> None:
        """

        :param excptn:
        :type excptn:"""
        self._pool.terminate()
        self._pool.close()
        # print(excptn.__cause__)
        # sys.exit(1)
        # exc_type, exc_obj, exc_tb = sys.exc_info()
        raise excptn

    def get(self) -> Any:
        """

        :return:"""
        if self.queue_size < 1:  # self._queue.empty():
            if len(multiprocessing.active_children()) == 0:
                if self.blocking:
                    self.maybe_fill()
                else:
                    raise StopIteration

        res = self._queue.get(self.blocking)
        self.maybe_fill()
        return res

    def __len__(self) -> int:
        return self.queue_size

    def __iter__(self):
        return self

    def __next__(self) -> Any:
        return self.get()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._pool.terminate()
        self._pool.close()
        if exc_type:
            # print(exc_type, exc_val, exc_tb) # trace_back
            raise exc_type(exc_val)
            # sys.exit()


if __name__ == "__main__":

    class Square(PooledQueueTask):
        def call(self, i, *args, **kwargs):
            """

            :param i:
            :type i:
            :param args:
            :type args:
            :param kwargs:
            :type kwargs:
            :return:
            :rtype:"""
            return i * 2

    class Exc(PooledQueueTask):
        def call(self, *args, **kwargs):
            """

            :param args:
            :type args:
            :param kwargs:
            :type kwargs:"""
            raise NotImplementedError

    task = Square()

    processor = PooledQueueProcessor(
        task, [2], fill_at_construction=True, max_queue_size=100
    )
    for GPU_STATS, _ in zip(processor, range(30)):
        print(GPU_STATS)

    processor.blocking = True
    processor.args = [4]
    time.sleep(3)
    for GPU_STATS in processor:
        print(GPU_STATS)
        if GPU_STATS == 8:
            break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/04/2020
           """

__all__ = [
    "rgb_drop_alpha_hwc",
    "rgb_drop_alpha_batch_nhwc",
    "torch_vision_normalize_batch_nchw",
    "reverse_torch_vision_normalize_batch_nchw",
]

# from numba import jit

from warg.typing_extension import StrictNumbers


# @jit(nopython=True, fastmath=True)
def rgb_drop_alpha_hwc(inp: StrictNumbers) -> StrictNumbers:
    """

    :param inp:
    :type inp:
    :return:
    :rtype:"""
    assert len(inp[-1, -1]) >= 3, f"not enough channels, only had {len(inp[-1, -1])}"
    return inp[..., :3]


# @jit(nopython=True, fastmath=True)
def rgb_drop_alpha_batch_nhwc(inp: StrictNumbers) -> StrictNumbers:
    """

    :param inp:
    :type inp:
    :return:
    :rtype:"""
    assert (
        len(inp[-1, -1, -1]) >= 3
    ), f"not enough channels, only had {len(inp[-1, -1, -1])}"
    return inp[..., :3]


# @jit(nopython=True, fastmath=True)
def torch_vision_normalize_batch_nchw(inp: StrictNumbers) -> StrictNumbers:
    """

      warning INPLACE!

    :param inp:
    :type inp:
    :return:
    :rtype:"""
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)
    assert len(inp[-1]) == 3, f"was {len(inp[-1])}"

    inp[:, 0] = (inp[:, 0] - mean[0]) / std[0]
    inp[:, 1] = (inp[:, 1] - mean[1]) / std[1]
    inp[:, 2] = (inp[:, 2] - mean[2]) / std[2]

    return inp


# @jit(nopython=True, fastmath=True)
def reverse_torch_vision_normalize_batch_nchw(inp: StrictNumbers) -> StrictNumbers:
    """

    :param inp:
    :type inp:
    :return:
    :rtype:"""
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)
    assert len(inp[-1]) == 3, f"was {len(inp[-1])}"

    inp[:, 0] = inp[:, 0] * std[0] + mean[0]
    inp[:, 1] = inp[:, 1] * std[1] + mean[1]
    inp[:, 2] = inp[:, 2] * std[2] + mean[2]

    return inp


if __name__ == "__main__":
    import numpy

    def asda() -> None:
        """
        :rtype: None
        """
        a = numpy.ones((1, 4, 4, 4))
        b = numpy.ones((1, 4, 4, 3))
        c = numpy.ones((4, 4, 3))
        d = numpy.ones((1, 4, 4, 2))

        rgb_drop_alpha_batch_nhwc(a)
        rgb_drop_alpha_batch_nhwc(b)
        try:
            rgb_drop_alpha_batch_nhwc(c)
        except:
            pass
        rgb_drop_alpha_hwc(c)
        try:
            rgb_drop_alpha_batch_nhwc(d)
        except:
            pass
        try:
            rgb_drop_alpha_hwc(d)
        except:
            pass

    def asbsdfdsa() -> None:
        """
        :rtype: None
        """
        a = numpy.ones((1, 3, 4, 4))
        ba = torch_vision_normalize_batch_nchw(a)
        print(ba)
        ca = reverse_torch_vision_normalize_batch_nchw(ba)
        print(ca)

    asda()
    asbsdfdsa()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 27-01-2021
           """

__all__ = [
    "max_decimation_subsample",
    "min_decimation_subsample",
    "mag_decimation_subsample",
    "fir_subsample",
    "fft_subsample",
]

from typing import Iterable, Tuple, Union

import numpy
from scipy.signal import decimate, resample


def max_decimation_subsample(
    signal: Union[Iterable, numpy.ndarray],
    decimation_factor: int = 10,
    return_indices: bool = False,
    truncate_last_indivisible: bool = True,
) -> numpy.ndarray:
    """ """
    signal = numpy.array(signal)
    if truncate_last_indivisible:
        div = len(signal) // decimation_factor
    signal = signal[: div * decimation_factor]

    s = signal.reshape(-1, decimation_factor)
    if return_indices:
        a = numpy.argmax(s, axis=1)
        return numpy.array(
            [ao * decimation_factor + am for ao, am in zip(range(len(a)), a)]
        )
    return numpy.max(s, axis=1)


def min_decimation_subsample(
    signal: Union[Iterable, numpy.ndarray],
    decimation_factor: int = 10,
    return_indices: bool = False,
    truncate_last_indivisible: bool = True,
) -> numpy.ndarray:
    """ """
    signal = numpy.array(signal)
    if truncate_last_indivisible:
        div = len(signal) // decimation_factor
    signal = signal[: div * decimation_factor]

    s = signal.reshape(-1, decimation_factor)
    if return_indices:
        a = numpy.argmin(s, axis=1)
        return numpy.array(
            [ao * decimation_factor + am for ao, am in zip(range(len(a)), a)]
        )
    return numpy.min(s, axis=1)


def mag_decimation_subsample(
    signal: Union[Iterable, numpy.ndarray],
    decimation_factor: int = 10,
    return_indices: bool = False,
    truncate_last_indivisible: bool = True,
) -> numpy.ndarray:
    """

    truncate_last_undivisible is false, signal be divisible by the decimation_factor
    """
    signal = numpy.array(signal)
    if truncate_last_indivisible:
        div = len(signal) // decimation_factor
        signal = signal[: div * decimation_factor]

    s = signal.reshape(-1, decimation_factor)
    s_min = numpy.argmin(s, axis=1)
    s_mi = [ao * decimation_factor + am for ao, am in zip(range(len(s_min)), s_min)]
    s_max = numpy.argmax(s, axis=1)
    s_ma = [ao * decimation_factor + am for ao, am in zip(range(len(s_max)), s_max)]
    s_mag = [
        smax if (numpy.abs(signal[smin]) < numpy.abs(signal[smax])) else smin
        for smin, smax in zip(s_mi, s_ma)
    ]
    if return_indices:
        return numpy.array(s_mag)
    return signal[s_mag]


def grad_decimation_subsample() -> None:
    """
    :rtype: None
    """
    # Gradient based windowed subsampling
    raise NotImplemented()


def fir_subsample(
    signal: numpy.ndarray, max_resolution: int, sampling_rate: int
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """ """
    if signal.shape[-1] > max_resolution:
        sub_signal = decimate(signal, (signal.shape[-1] // max_resolution) + 1, axis=-1)
    else:
        sub_signal = signal
    sub_time = numpy.linspace(
        0,
        signal.shape[-1] // sampling_rate,  # Get time from indices
        num=sub_signal.shape[-1],
    )
    return sub_time, sub_signal


def fft_subsample(
    signal: numpy.ndarray, max_resolution: int, sampling_rate: int
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """ """
    if signal.shape[-1] > max_resolution:
        sub_signal = resample(signal, max_resolution, axis=-1)
    else:
        sub_signal = signal
    sub_time = numpy.linspace(
        0,
        signal.shape[-1] // sampling_rate,  # Get time from indices
        num=sub_signal.shape[-1],
    )
    return sub_time, sub_signal


if __name__ == "__main__":

    def aisjd() -> None:
        """
        :rtype: None
        """
        t = numpy.linspace(-1, 1, 200)
        a = (
            numpy.sin(2 * numpy.pi * 0.75 * t * (1 - t) + 2.1)
            + 0.1 * numpy.sin(2 * numpy.pi * 1.25 * t + 1)
            + 0.18 * numpy.cos(2 * numpy.pi * 3.85 * t)
        )
        print(a)
        print(max_decimation_subsample(a, decimation_factor=10, return_indices=True))
        print(min_decimation_subsample(a, decimation_factor=10, return_indices=True))
        print(mag_decimation_subsample(a, decimation_factor=10, return_indices=True))

    aisjd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math
from typing import Iterable, Iterator, Sequence, Tuple

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

from matplotlib.figure import Figure
from matplotlib.pyplot import matshow, imshow
from itertools import cycle

from numpy import interp

from warg import passes_kws_to
from matplotlib import pyplot
import numpy
from sklearn import svm
from sklearn.decomposition import PCA
from sklearn.metrics import (
    auc,
    average_precision_score,
    confusion_matrix,
    precision_recall_curve,
    roc_curve,
)
from sklearn.model_selection import train_test_split
from sklearn.multiclass import OneVsRestClassifier
from sklearn.preprocessing import StandardScaler, label_binarize
from sklearn.utils.multiclass import unique_labels

__all__ = [
    "horizontal_imshow",
    "biplot",
    "pca_biplot",
    "correlation_matrix_plot",
    "confusion_matrix_plot",
    "roc_plot",
    "precision_recall_plot",
]


@passes_kws_to(matshow)
def correlation_matrix_plot(
    cor: Sequence, labels: Sequence = None, title: str = "", **kwargs
) -> Figure:
    """

    :param cor:
    :type cor:
    :param labels:
    :type labels:
    :param title:
    :type title:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """
    if labels is None:
        labels = [f"P{i}" for i in range(len(cor))]

    fig = pyplot.figure()
    ax1 = fig.add_subplot(111)

    cax = ax1.matshow(cor, **kwargs)
    ax1.grid(True)
    pyplot.title(title)

    ax1.set_xticks(range(len(cor)))
    ax1.set_yticks(range(len(cor)))

    ax1.set_xticklabels(labels, fontsize=6)
    ax1.set_yticklabels(labels, fontsize=6)

    fig.colorbar(cax, ticks=numpy.arange(-1.1, 1.1, 0.1))

    return fig


@passes_kws_to(imshow)
def horizontal_imshow(
    images: Sequence, titles: Sequence = None, num_columns: int = 4, **kwargs
):
    """Small helper function for creating horizontal subplots with pyplot

    :param images:
    :param titles:
    :param num_columns:
    :param kwargs:
    :return:
    """
    num_d = len(images) / num_columns
    num_d_f = math.floor(num_d)
    if num_d_f != num_d:
        num_d_f += 1

    if titles is None:
        titles = [f"fig{a_}" for a_ in range(len(images))]
    fig, axes = pyplot.subplots(
        num_d_f,
        num_columns,
        squeeze=False,
        sharex="all",
        sharey="all",
        constrained_layout=True,
    )
    figure_axes = []
    for a_ in axes:
        figure_axes.extend(a_)
    for i, image in enumerate(images):
        ax = figure_axes[i]
        if titles:
            ax.set_title(titles[i])
        ax.imshow(image, **kwargs)
    return fig


def biplot(
    scores: Sequence,
    coefficients: numpy.ndarray,
    categories: Iterable = None,
    labels: Sequence = None,
    label_multiplier: float = 1.06,
) -> pyplot.Figure:
    """
       Use only the 2 Principal components.

      :rtype: object
    :param label_multiplier:
    :param categories:
    :param scores:
    :param coefficients:
    :param labels:
    :return:"""
    assert label_multiplier >= 1.0

    fig = pyplot.figure()

    ax1 = fig.add_axes([0.12, 0.1, 0.76, 0.8], label="ax1")
    ax2 = fig.add_axes([0.12, 0.1, 0.76, 0.8], label="ax2", frameon=False)

    ax1.yaxis.tick_left()
    ax1.xaxis.tick_bottom()

    ax2.yaxis.tick_right()
    ax2.yaxis.set_label_position("right")
    ax2.yaxis.set_offset_position("right")
    ax2.xaxis.tick_top()
    ax2.xaxis.set_label_position("top")

    ax1.spines["right"].set_color("red")
    ax1.spines["top"].set_color("red")

    ax2.tick_params(color="red")

    for ylabel, xlabel in zip(ax2.get_yticklabels(), ax2.get_xticklabels()):
        ylabel.set_color("red")
        xlabel.set_color("red")

    ax1.set_xlabel("$Score_{PC1}$")
    ax1.set_ylabel("$Score_{PC2}$")

    ax2.set_xlabel("$Coefficient_{PC1}$", color="red")
    ax2.set_ylabel("$Coefficient_{PC2}$", color="red")

    ax2.set_xlim(-label_multiplier, label_multiplier)
    ax2.set_ylim(-label_multiplier, label_multiplier)

    xs = scores[:, 0]
    ys = scores[:, 1]
    n = coefficients.shape[0]
    scale_x = 1.0 / (xs.max() - xs.min())
    scale_y = 1.0 / (ys.max() - ys.min())
    ax1.scatter(xs * scale_x, ys * scale_y, c=categories)
    # pyplot.colorbar(ax=ax1)

    for i in range(n):
        ax2.arrow(0, 0, coefficients[i, 0], coefficients[i, 1], color="r", alpha=0.5)

        label = f"$x_{i}$"
        if labels is not None:
            label = labels[i]

        ax2.text(
            coefficients[i, 0] * label_multiplier,
            coefficients[i, 1] * label_multiplier,
            label,
            color="g",
            ha="center",
            va="center",
        )

    return fig


def pca_biplot(
    predictor: Iterable, response: Iterable, labels: Iterable[str] = None
) -> pyplot.Figure:
    """
    produces a pca projection and plot the 2 most significant component score and the component coefficients.

    :param predictor:
    :param response:
    :param labels:
    :return:"""

    scaler = StandardScaler()
    scaler.fit(predictor)

    pca = PCA()

    return biplot(
        pca.fit_transform(scaler.transform(predictor))[:, 0:2],
        numpy.transpose(pca.components_[0:2, :]),
        response,
        labels,
    )


def precision_recall_plot(
    truth,
    score,
    num_classes,
    *,
    num_decimals: int = 2,
    include_thresholds: bool = False,
    y_lim: Tuple[float, float] = (0.0, 1.05),
    figure_size: Tuple[int, int] = (7, 9),
    color_cycle: Iterator = cycle(
        ["navy", "turquoise", "darkorange", "cornflowerblue", "teal"]
    ),
) -> pyplot.Figure:
    """

    # A "micro-average": quantifying score on all classes jointly

    :param truth:
    :param score:
    :param num_classes:
    :param num_decimals:
    :param include_thresholds:
    :param y_lim:
    :param figure_size:
    :param color_cycle:
    :return:
    """
    precision = dict()
    recall = dict()
    average_precision = dict()
    thresholds = dict()
    for i in range(num_classes):
        precision[i], recall[i], thresholds[i] = precision_recall_curve(
            truth[:, i], score[:, i]
        )
        average_precision[i] = average_precision_score(truth[:, i], score[:, i])

    precision["micro"], recall["micro"], thresholds["micro"] = precision_recall_curve(
        truth.ravel(), score.ravel()
    )
    average_precision["micro"] = average_precision_score(truth, score, average="micro")

    fig = pyplot.figure(figsize=figure_size)
    f_scores = numpy.linspace(0.2, 0.8, num=4)
    lines = []
    labels = []
    iso_curves = None
    for f_score in f_scores:
        x = numpy.linspace(0.01, 1)
        y = f_score * x / (2 * x - f_score)
        iso_curves = pyplot.plot(x[y >= 0], y[y >= 0], color="gray", alpha=0.2)
        pyplot.annotate(f"f1={f_score:0.1f}", xy=(0.9, y[45] + 0.02))

    lines.extend(iso_curves)
    labels.append("iso-f1 curves")

    lines.extend(pyplot.plot(recall["micro"], precision["micro"], color="gold", lw=2))
    if include_thresholds:
        for p, r, t in zip(precision["micro"], recall["micro"], thresholds["micro"]):
            pyplot.annotate(f"{t:.{num_decimals}f}", (r, p))
    labels.append(
        f"micro-average Precision-recall (area = {average_precision['micro']:0.{num_decimals}f})"
    )

    for i, color in zip(range(num_classes), color_cycle):
        lines.extend(pyplot.plot(recall[i], precision[i], color=color, lw=2))
        if include_thresholds:
            for p, r, t in zip(precision[i], recall[i], thresholds[i]):
                pyplot.annotate(f"{t:.{num_decimals}f}", (r, p))
        labels.append(
            f"Precision-recall for class {i} (area = {average_precision[i]:0.{num_decimals}f})"
        )

    pyplot.plot([0, 1], [0.5, 0.5], linestyle="--")  # plot no skill

    fig = pyplot.gcf()
    fig.subplots_adjust(bottom=0.25)
    pyplot.xlim([0.0, 1.0])
    pyplot.ylim(y_lim)
    pyplot.xlabel("Recall")
    pyplot.ylabel("Precision")
    pyplot.title("Multi-class extension of Precision-Recall curve")
    pyplot.legend(lines, labels, loc=(0, -0.38), prop=dict(size=14))

    return fig


def roc_plot(
    truth,
    score,
    num_classes: int,
    *,
    figure_size: Tuple[int, int] = (8, 8),
    num_decimals: int = 2,
) -> Figure:
    """ """
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(num_classes):
        fpr[i], tpr[i], _ = roc_curve(truth[:, i], score[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(truth.ravel(), score.ravel())
    roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

    # Compute macro-average ROC curve and ROC area
    all_fpr = numpy.unique(
        numpy.concatenate([fpr[i] for i in range(num_classes)])
    )  # First aggregate all false positive rates
    mean_tpr = numpy.zeros_like(
        all_fpr
    )  # Then interpolate all ROC curves at this points
    for i in range(num_classes):
        mean_tpr += interp(all_fpr, fpr[i], tpr[i])
    mean_tpr /= num_classes  # Finally average it and compute AUC
    fpr["macro"] = all_fpr
    tpr["macro"] = mean_tpr
    roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

    # Plot all ROC curves
    fig = pyplot.figure(figsize=figure_size)
    lw = 1
    pyplot.plot([0, 1], [0, 1], "k--", lw=lw)

    pyplot.plot(
        fpr["micro"],
        tpr["micro"],
        label=f'micro-average ROC curve (area = {roc_auc["micro"]:0.{num_decimals}f})',
        color="deeppink",
        linestyle=":",
        linewidth=lw,
    )

    pyplot.plot(
        fpr["macro"],
        tpr["macro"],
        label=f'macro-average ROC curve (area = {roc_auc["macro"]:0.{num_decimals}f})',
        color="navy",
        linestyle=":",
        linewidth=lw,
    )

    colors = cycle(["aqua", "darkorange", "cornflowerblue", "red", "green", "teal"])
    for i, color in zip(range(num_classes), colors):
        pyplot.plot(
            fpr[i],
            tpr[i],
            color=color,
            lw=lw,
            label=f"ROC curve of class {i} (area = {roc_auc[i]:0.{num_decimals}f})",
        )

    pyplot.xlim([-0.05, 1.0])
    pyplot.ylim([0.0, 1.05])
    pyplot.xlabel("False Positive Rate")
    pyplot.ylabel("True Positive Rate")
    pyplot.title("Multi-class Extension Receiver Operating Characteristic")
    pyplot.legend(loc="lower right")

    return fig


def confusion_matrix_plot(
    truth: Iterable,
    prediction: Iterable,
    category_names: Iterable[str],
    *,
    figure_size: Tuple[int, int] = (8, 8),
    decimals: int = 3,
) -> Figure:
    """

    :param truth:
    :type truth:
    :param prediction:
    :type prediction:
    :param category_names:
    :type category_names:
    :param figure_size:
    :type figure_size:
    :param decimals:
    :type decimals:
    :return:
    :rtype:"""

    def confusion_matrix_figure(
        truth_, prediction_, classes, normalize=False, title=None, cmap=pyplot.cm.Blues
    ):
        """
        This function prints and plots the confusion matrix.
        Normalization can be applied by setting `normalize=True`."""
        if not title:
            if normalize:
                title = "Normalized confusion matrix"
            else:
                title = "Confusion matrix, without normalization"

        cm = confusion_matrix(truth_, prediction_)
        unique = unique_labels(
            truth_, prediction_
        )  # Only use the labels that appear in the data
        classes = classes[unique]
        if normalize:
            cm = cm.astype("float") / cm.sum(axis=1)[:, numpy.newaxis]

        fig, ax = pyplot.subplots(figsize=figure_size)
        im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
        pyplot.colorbar(im, ax=ax)

        ax.set(
            xticks=numpy.arange(cm.shape[1]),
            yticks=numpy.arange(cm.shape[0]),
            # ... and label them with the respective list entries
            xticklabels=classes,
            yticklabels=classes,
            title=title,
            ylabel="True label",
            xlabel="Predicted label",
        )  # Show all ticks

        pyplot.setp(
            ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor"
        )  # Rotate the tick labels and set their alignment.

        # Loop over data dimensions and create text annotations.
        fmt = ".2f" if normalize else "d"
        thresh = cm.max() / 2.0
        for i in range(cm.shape[0]):
            for j in range(cm.shape[1]):
                ax.text(
                    j,
                    i,
                    format(cm[i, j], fmt),
                    ha="center",
                    va="center",
                    color="white" if cm[i, j] > thresh else "black",
                )
        fig.tight_layout()
        return fig

    numpy.set_printoptions(precision=decimals)

    # Plot normalized confusion matrix
    return confusion_matrix_figure(
        truth,
        prediction,
        classes=category_names,
        normalize=True,
        title="Normalized confusion matrix",
    )


if __name__ == "__main__":

    def a() -> None:
        """
        :rtype: None
        """
        from sklearn import datasets

        iris = datasets.load_iris()
        x = iris.data
        y = iris.target

        num_classes = len(iris.target_names)

        # Add noisy features to make the problem harder
        random_state = numpy.random.RandomState(0)
        n_samples, n_features = x.shape
        # X_noisy = numpy.c_[X, random_state.randn(n_samples, 200 * n_features)*0.01]

        x_noisy = x

        (X_train, X_test, y_train, y_test) = train_test_split(
            x_noisy,
            label_binarize(y, classes=range(len(iris.target_names))),
            test_size=0.3,
            random_state=0,
            shuffle=True,
        )  # shuffle and split training and test sets

        classifier = OneVsRestClassifier(
            svm.SVC(kernel="linear", probability=True, random_state=random_state)
        )  # Learn to predict each class against the other
        y_score = classifier.fit(X_train, y_train).decision_function(X_test)

        def sasasasgasgssiasjdijasagsaagdi():
            """ """
            confusion_matrix_plot(
                numpy.argmax(y_test, axis=-1),
                numpy.argmax(y_score, axis=-1),
                category_names=iris.target_names,
            )
            pyplot.show()

        def sasasasgasgssagsaagdi():
            """ """
            roc_plot(y_test, y_score, num_classes)
            pyplot.show()

        def sasasgssagsaagdi():
            """ """
            precision_recall_plot(y_test, y_score, num_classes)
            pyplot.show()

        def sasafgsagdi():
            """ """
            pca_biplot(x, y)
            pyplot.show()

        def sadi():
            """ """
            import pandas

            df = pandas.DataFrame(x, columns=iris.feature_names)
            correlation_matrix_plot(cor=df.corr(), labels=df.columns)
            pyplot.show()

        sasasasgasgssiasjdijasagsaagdi()
        sasasasgasgssagsaagdi()
        sasasgssagsaagdi()
        sasafgsagdi()
        sadi()

    a()
